# -*- coding: utf-8 -*-
"""This module injects some additional methods into owlready2 classes."""
from types import MethodType

import owlready2
from owlready2 import ThingClass, PropertyClass, Thing, Restriction


# Improve default rendering of entities
def render_func(entity):
    if hasattr(entity, 'prefLabel') and entity.prefLabel:
        name = entity.prefLabel[0]
    elif hasattr(entity, 'label') and entity.label:
        name = entity.label[0]
    elif hasattr(entity, 'altLabel') and entity.altLabel:
        name = entity.altLabel[0]
    else:
        name = entity.name
    return "%s.%s" % (entity.namespace.name, name)


owlready2.set_render_func(render_func)



#
# Extending ThingClass (classes)
#
def get_preferred_label(self):
    """Returns the preferred label as a string (not list).

    The following heuristics is used:
      - if prefLabel annotation property exists, returns the first prefLabel
      - if label annotation property exists, returns the first label
      - otherwise return the name
    """
    if hasattr(self, 'prefLabel') and self.prefLabel:
        return self.prefLabel[0]
    elif hasattr(self, 'label') and self.label:
        return self.label.first()
    else:
        return self.name


def get_parents(self, strict=False):
    """Returns a list of all parents.  If `strict` is true, parents that are
    parents of other parents are excluded."""
    if strict:
        s = self.get_parents()
        for e in s.copy():
            s.difference_update(e.ancestors(include_self=False))
        return s
    elif isinstance(self, ThingClass):
        return {cls for cls in self.is_a
                if isinstance(cls, ThingClass)}
    elif isinstance(self, owlready2.ObjectPropertyClass):
        return {cls for cls in self.is_a
                if isinstance(cls, owlready2.ObjectPropertyClass)}
    else:
        assert 0


def _dir(self):
    """Extend in dir() listing of ontology classes."""
    s = set(object.__dir__(self))
    props = self.namespace.world._props.keys()
    s.update(props)
    return sorted(s)


def get_class_annotations(self, all=False, imported=True):
    """Returns a dict with non-empty annotations.

    If `all` is true, also annotations with no value are included.

    If `imported` is true, also include annotations defined in
    imported ontologies.
    """
    onto = self.namespace.ontology
    d = {get_preferred_label(a): a._get_values_for_class(self)
         for a in onto.annotation_properties(imported=imported)}
    if all:
        return d
    else:
        return {k: v for k, v in d.items() if v}


def disjoint_with(self, reduce=False):
    """Returns a generator with all classes that are disjoint with `self`.
    If `reduce` is true, all classes that are a descendant of another class
    will be excluded."""
    if reduce:
        s = set(self.disjoint_with())
        for e in s.copy():
            s.difference_update(e.descendants(include_self=False))
        for e in s:
            yield e
    else:
        for d in self.disjoints():
            for e in d.entities:
                if e is not self:
                    yield e


def get_indirect_is_a(self, skip_classes=True):
    """Returns the set of all isSubclassOf relations of self and its
    ancestors.  If `skip_classes` is true, indirect classes are not
    included in the returned set.
    """
    s = set()
    for e in reversed(self.mro()):
        if hasattr(e, 'is_a'):
            if skip_classes:
                s.update(r for r in e.is_a
                         if not isinstance(r, owlready2.ThingClass))
            else:
                s.update(e.is_a)
    s.update(self.is_a)
    return s


#
# Extending PropertyClass (properties)
#
def get_property_annotations(self, all=False, imported=True):
    """Returns a dict with non-empty property annotations.

    If `all` is true, also annotations with no value are included.

    If `imported` is true, also include annotations defined in
    imported ontologies.
    """
    onto = self.namespace.ontology
    d = {get_preferred_label(a): a._get_values_for_class(self)
         for a in onto.annotation_properties(imported=imported)}
    if all:
        return d
    else:
        return {k: v for k, v in d.items() if v}


#
# Extending Thing (individuals)
#
def get_individual_annotations(self, all=False, imported=True):
    """Returns a dict with non-empty individual annotations.

    If `all` is true, also annotations with no value are included.

    If `imported` is true, also include annotations defined in
    imported ontologies.
    """
    onto = self.namespace.ontology
    d = {get_preferred_label(a): a._get_values_for_individual(self)
         for a in onto.annotation_properties(imported=imported)}
    if all:
        return d
    else:
        return {k: v for k, v in d.items() if v}


#
# Extending Restruction
#
def get_typename(self):
    return owlready2.class_construct._restriction_type_2_label[self.type]


# Inject methods into Owlready2 classes
setattr(ThingClass, '__dir__', _dir)
setattr(ThingClass, 'get_preferred_label', get_preferred_label)
setattr(ThingClass, 'get_parents', get_parents)
setattr(ThingClass, 'get_annotations', get_class_annotations)
setattr(ThingClass, 'disjoint_with', disjoint_with)
setattr(ThingClass, 'get_indirect_is_a', get_indirect_is_a)

setattr(PropertyClass, 'get_preferred_label', get_preferred_label)
setattr(PropertyClass, 'get_parents', get_parents)
setattr(PropertyClass, 'get_annotations', get_property_annotations)

setattr(Restriction, 'get_typename', get_typename)

# Method names for individuals must be different from method names for classes
type.__setattr__(Thing, 'get_preflabel', get_preferred_label)
type.__setattr__(Thing, 'get_individual_annotations', get_individual_annotations)



#!/usr/bin/env python3
import sys
import os

# Add emmo to sys path
thisdir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(1, os.path.abspath(os.path.join(thisdir, '..', '..')))
from emmo import get_ontology  # noqa: E402, F401


emmo = get_ontology()
emmo.load()

graph = emmo.get_dot_graph(relations='is_a')
graph.write_svg('taxonomy.svg')
graph.write_pdf('taxonomy.pdf')

entity_graph = emmo.get_dot_graph('EMMO')
entity_graph.write_svg('taxonomy2.svg')

substrate_graph = emmo.get_dot_graph('Item', relations=True,
                                     leafs=('Physical'), parents='Item',
                                     style='uml')
substrate_graph.write_svg('merotopology_graph.svg')

property_graph = emmo.get_dot_graph('Property')
property_graph.write_svg('property_graph.svg')


# Update the default style
emmo._default_style['graph']['rankdir'] = 'BT'

relations_graph = emmo.get_dot_graph('EMMORelation')
relations_graph.write_pdf('relation_graph.pdf')
relations_graph.write_png('relation_graph.png')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""\
Step 2 - define metadata for the ASE Atoms class
------------------------------------------------
In this step we define metadata for the Atoms class in the Atomistic
Simulation Environment (ASE).  This metadata is defined in the file
atoms.json.

The actual definition are provided in atoms.json.  Below we use the
dlite.classfactory() to create a subclass of ASE Atoms that also
exposes the attributes as dlite properties. The subclass DLiteAtoms
adds some methods for handling some special attributes.

"""
import ase
from ase.spacegroup import Spacegroup

import dlite


# Create an ASE Atoms subclass that also inherits from dlite atoms.json
BaseAtoms = dlite.classfactory(ase.Atoms, url='json://atoms.json?mode=r#')


class DLiteAtoms(BaseAtoms):
    """ASE Atoms class extended as a dlite entity."""

    def _dlite_get_info(self):
        d = self.info.copy()
        sg = Spacegroup(d.get('spacegroup', 'P 1'))
        d['spacegroup'] = sg.symbol
        return [(k, str(v)) for k, v in d.items()]

    def _dlite_set_info(self, value):
        self.info.update(value)
        self.info['spacegroup'] = Spacegroup(self.info['spacegroup'])

    def _dlite_get_celldisp(self):
        return self.get_celldisp()[:, 0]



#!/usr/bin/env python3
import sys
import os

# Add emmo to sys path
thisdir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(1, os.path.abspath(os.path.join(thisdir, '..', '..')))
from emmo import get_ontology  # noqa: E402, F401
from emmo.graph import OntoGraph  # noqa: E402, F401

# Create output directory
outdir = 'test_graph2'
if not os.path.exists(outdir):
    os.makedirs(outdir)
os.chdir(outdir)


emmo = get_ontology()
emmo.load()


g = OntoGraph(emmo, emmo.hasPart, leafs=('mereotopological', 'semiotical',
                                         'connected'))
g.save('hasPart.svg')


g = OntoGraph(emmo, emmo.Matter, relations='all', addnodes=True,
              edgelabels=None)
g.save('Matter.png')

g = OntoGraph(emmo, emmo.ElementaryParticle, relations='all', addnodes=True,
              edgelabels=None)
g.save('ElementaryParticle.png')

g = OntoGraph(emmo, emmo.SIBaseUnit, relations='all', addnodes=True,
              edgelabels=True,
              addconstructs=False, graph_attr={'rankdir': 'RL'})
g.add_legend()
g.save('SIBaseUnit.png')

g = OntoGraph(emmo, emmo.EMMORelation, relations='all', edgelabels=None)
g.save('EMMORelation.png')

g = OntoGraph(emmo, emmo.Quantity,
              leafs=[emmo.DerivedQuantity, emmo.BaseQuantity,
                     emmo.PhysicalConstant],
              relations='all', edgelabels=None, addnodes=True,
              addconstructs=True, graph_attr={'rankdir': 'RL'})
g.add_legend()
g.save('Quantity.svg')


# Used for figures

g = OntoGraph(emmo)
g.add_legend('all')
g.save('legend.png')


g = OntoGraph(emmo, emmo.EMMO, leafs=[emmo.Perspective, emmo.Elementary])
g.save('top.png')


leafs = set()
for s in emmo.Perspective.subclasses():
    leafs.update(s.subclasses())
g = OntoGraph(emmo, emmo.Perspective, leafs=leafs, parents=1)
g.save('Perspectives.png')


leafs = {emmo.Interpreter, emmo.Conventional, emmo.Icon, emmo.Observation,
         emmo.Object}
hidden = {emmo.SIUnitSymbol, emmo.SpecialUnit, emmo.Manufacturing,
          emmo.Engineered, emmo.PhysicalPhenomenon}
semiotic = emmo.get_branch(emmo.Holistic, leafs=leafs.union(hidden))
semiotic.difference_update(hidden)
g = OntoGraph(emmo)
g.add_entities(semiotic, relations='all', edgelabels=False)
g.save('Semiotic.png')
g.add_legend()
g.save('Semiotic+legend.png')

legend = OntoGraph(emmo)
legend.add_legend(g.get_relations())
legend.save('Semiotic-legend.png')


# Measurement
leafs = {emmo.Object}

hidden = {emmo.SIUnitSymbol, emmo.SpecialUnit, emmo.Manufacturing,
          emmo.Engineered, emmo.PhysicalPhenomenon,
          emmo.Icon, emmo.Interpretant, emmo.Index,
          emmo.SubjectiveProperty,
          emmo.NominalProperty,
          emmo.ConventionalQuantitativeProperty,
          emmo.ModelledQuantitativeProperty,
          emmo.Theorization, emmo.Experiment, emmo.Theory, emmo.Variable}
semiotic = emmo.get_branch(emmo.Holistic, leafs=leafs.union(hidden))
semiotic.difference_update(hidden)
g = OntoGraph(emmo)
g.add_entities(semiotic, relations='all', edgelabels=False)
g.add_legend()
g.save('measurement.png')



# Reductionistic perspective
g = OntoGraph(emmo, emmo.Reductionistic, relations='all', addnodes=False,
              leafs=[emmo.Quantity, emmo.String, emmo.PrefixedUnit,
                     emmo.SymbolicComposition, emmo.Matter],
              parents=2,
              #entities=[emmo.Symbolic],
              edgelabels=None)
g.add_legend()
g.save('Reductionistic.png')



# -*- coding: utf-8 -*-
"""A module adding additional functionality to owlready2. The main additions
includes:
  - Visualisation of taxonomy and ontology as graphs (using pydot, see
    ontograph.py).

The class extension is defined within.

If desirable some of this may be moved back into owlready2.
"""
import sys
import os
import itertools
import inspect
import warnings
import uuid
import tempfile
from collections import defaultdict

from rdflib.util import guess_format
import owlready2

from .utils import asstring, read_catalog, infer_version, convert_imported
from .ontograph import OntoGraph  # FIXME: depricate...


class NoSuchLabelError(LookupError, AttributeError):
    """Error raised when a label cannot be found."""
    pass


# owl categories
categories = (
    'annotation_properties',
    'data_properties',
    'object_properties',
    'classes',
    'individuals',
)


def get_ontology(*args, **kwargs):
    """Returns a new Ontology from `base_iri`.

    This is a convenient function for calling World.get_ontology()."""
    return World().get_ontology(*args, **kwargs)


def isinteractive():
    """Returns true if we are running from an interactive interpreater,
    false otherwise."""
    return bool(hasattr(__builtins__, '__IPYTHON__') or
                sys.flags.interactive or
                hasattr(sys, 'ps1'))


class World(owlready2.World):
    """A subclass of owlready2.World."""

    def get_ontology(self, base_iri='emmo-inferred'):
        """Returns a new Ontology from `base_iri`.

        The `base_iri` argument may be one of:
          - valid URL (possible excluding final .owl)
          - file name (possible excluding final .owl)
          - "emmo": load latest stable version of asserted EMMO
          - "emmo-inferred": load latest stable version of inferred EMMO
            (default)
          - "emmo-development": load latest inferred development version
            of EMMO
        """
        if base_iri == 'emmo':
            base_iri = 'http://emmo.info/emmo'
        elif base_iri == 'emmo-inferred':
            base_iri = (
                'https://emmo-repo.github.io/latest-stable/emmo-inferred.owl')
        elif base_iri == 'emmo-development':
            base_iri = (
                'https://emmo-repo.github.io/development/emmo-inferred.owl')

        if base_iri in self.ontologies:
            onto = self.ontologies[base_iri]
        elif base_iri + '#' in self.ontologies:
            onto = self.ontologies[base_iri + '#']
        elif base_iri + '/' in self.ontologies:
            onto = self.ontologies[base_iri + '/']
        else:
            if os.path.exists(base_iri):
                iri = os.path.abspath(base_iri)
            elif os.path.exists(base_iri + '.owl'):
                iri = os.path.abspath(base_iri + '.owl')
            else:
                iri = base_iri

            if iri[-1] not in '/#':
                iri += '#'
            onto = Ontology(self, iri)

        return onto


class Ontology(owlready2.Ontology, OntoGraph):
    """A generic class extending owlready2.Ontology.
    """

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __getattr__(self, name):
        attr = super().__getattr__(name)
        if not attr:
            attr = self.get_by_label(name)
        return attr

    # Some properties for customising dir() listing.
    # Very useful in interactive sessions.
    dir_preflabel = property(
        fget=lambda self: getattr(self, '_dir_preflabel', isinteractive()),
        fset=lambda self, v: setattr(self, '_dir_preflabel', bool(v)),
        doc='Whether to include entity prefLabel in dir() listing.')
    dir_label = property(
        fget=lambda self: getattr(self, '_dir_label', isinteractive()),
        fset=lambda self, v: setattr(self, '_dir_label', bool(v)),
        doc='Whether to include entity label in dir() listing.')
    dir_name = property(
        fget=lambda self: getattr(self, '_dir_name', False),
        fset=lambda self, v: setattr(self, '_dir_name', bool(v)),
        doc='Whether to entity name in dir() listing.')
    dir_imported = property(
        fget=lambda self: getattr(self, '_dir_imported', isinteractive()),
        fset=lambda self, v: setattr(self, '_dir_imported', bool(v)),
        doc='Whether to include imported ontologies in dir() '
        'listing.')

    def __dir__(self):
        s = set(super().__dir__())
        if self.dir_preflabel:
            s.update(e.prefLabel.first() for e in
                     self.get_entities(imported=self.dir_imported)
                     if hasattr(e, 'prefLabel'))
        if self.dir_label:
            s.update(e.label.first() for e in
                     self.get_entities(imported=self.dir_imported)
                     if hasattr(e, 'label'))
        if self.dir_name:
            s.update(e.name for e in
                     self.get_entities(imported=self.dir_imported)
                     if hasattr(e, 'name'))
        s.difference_update({None})  # get rid of possible None
        return sorted(s)

    def __contains__(self, other):
        try:
            self[other]
            return True
        except NoSuchLabelError:
            return False

    def __objclass__(self):
        # Play nice with inspect...
        pass

    def load(self, only_local=False, filename=None, format=None,
             reload=None, reload_if_newer=False, url_from_catalog=False,
             catalog_file='catalog-v001.xml', tmpdir=None,
             **kwargs):
        """Load the ontology.

        Parameters
        ----------
        only_local : bool
            Whether to only read local files.  This requires that you
            have appended the path to the ontology to owlready2.onto_path.
        filename : str
            Path to file to load the ontology from.  Defaults to `base_iri`
            provided to get_ontology().
        format : str
            Format of `filename`.  Default is inferred from `filename`
            extension.
        reload : bool
            Whether to reload the ontology if it is already loaded.
        reload_if_newer : bool
            Whether to reload the ontology if the source has changed since
            last time it was loaded.
        url_from_catalog : bool
            Use catalog file if `base_iri` cannot be resolved.
        catalog_file : str
            Name of Protègè catalog file in the same folder as the
            ontology.  This option is used together with `only_local` and
            defaults to "catalog-v001.xml".
        tmpdir : str
            Path to temporary directory.
        kwargs
            Additional keyword arguments are passed on to
            owlready2.Ontology.load().
        """
        # If filename is not given, infer it from base_iri (if possible)
        web_protocols = ('http://', 'https://', )
        if not filename and not self.base_iri.startswith(web_protocols):
            filename = self.base_iri.rstrip('#/')
            if filename.startswith('file://'):
                filename = filename[7:]

        # Convert filename to owl if it is in a format not supported
        # by owlready2
        if filename:
            if not format:
                fmap = {'n3': 'ntriples'}
                format = guess_format(filename)
            if format not in ('xml', 'ntriples'):
                with tempfile.TemporaryDirectory() as tmpdir:
                    if not os.path.exists(tmpdir):
                        os.makedirs(tmpdir)
                    output = os.path.join(tmpdir, os.path.splitext(
                        os.path.basename(filename))[0] + '.owl')
                    convert_imported(filename, output,
                                     input_format=format, output_format='xml',
                                     catalog_file=catalog_file)
                    return self.load(only_local=True,
                                     filename=output,
                                     format='xml',
                                     reload=reload,
                                     reload_if_newer=reload_if_newer,
                                     url_from_catalog=url_from_catalog,
                                     catalog_file=catalog_file,
                                     **kwargs)

        # Append paths from catalog file to onto_path
        dirpath = os.path.normpath(
            os.path.dirname(filename or self.base_iri.rstrip('/#')))
        if only_local and os.path.exists(os.path.join(dirpath, catalog_file)):
            iris, dirs = read_catalog(
                dirpath, recursive=True, return_paths=True,
                catalog_file=catalog_file)
            for d in sorted(dirs, reverse=True):
                if d not in owlready2.onto_path:
                    owlready2.onto_path.append(d)

        fileobj = open(filename, 'rb') if filename else None

        try:
            super().load(only_local=only_local, fileobj=fileobj, reload=reload,
                         reload_if_newer=reload_if_newer, **kwargs)
        except owlready2.OwlReadyOntologyParsingError:
            if url_from_catalog:
                # Use catalog file to update IRIs of imported ontologies
                # in internal store and try to load again...
                iris = read_catalog(dirpath, catalog_file=catalog_file)
                for abbrev_iri in self.world._get_obj_triples_sp_o(
                        self.storid, owlready2.owl_imports):
                    iri = self._unabbreviate(abbrev_iri)
                    if iri in iris:
                        self._del_obj_triple_spo(self.storid,
                                                 owlready2.owl_imports,
                                                 abbrev_iri)
                        self._add_obj_triple_spo(self.storid,
                                                 owlready2.owl_imports,
                                                 self._abbreviate(iris[iri]))
                self.loaded = False
                super().load(only_local=only_local, fileobj=fileobj,
                             reload=reload, reload_if_newer=reload_if_newer,
                             **kwargs)
            else:
                raise
        finally:
            if fileobj:
                fileobj.close()

        return self

    def save(self, filename=None, format='rdfxml', overwrite=False, **kwargs):
        """Writes the ontology to file.

        If `overwrite` is true and filename exists, it will be removed
        before saving.  The default is to append an existing ontology.
        """
        if overwrite and filename and os.path.exists(filename):
            os.remove(filename)
        super().save(file=filename, format=format, **kwargs)

    def get_imported_ontologies(self, recursive=False):
        """Return a list with imported ontologies.

        If `recursive` is true, ontologies imported by imported ontologies
        are also returned.
        """
        def rec_imported(onto):
            for o in onto.imported_ontologies:
                imported.add(o)
                rec_imported(o)

        if recursive:
            imported = set()
            rec_imported(self)
            return list(imported)
        else:
            return self.imported_ontologies

    def get_entities(self, imported=True, classes=True, individuals=True,
                     object_properties=True, data_properties=True,
                     annotation_properties=True):
        """Return a generator over (optionally) all classes, individuals,
        object_properties, data_properties and annotation_properties.

        If `imported` is true, entities in imported ontologies will also
        be included.
        """
        categories = set([
            'classes' if classes else None,
            'individuals' if individuals else None,
            'object_properties' if object_properties else None,
            'data_properties' if data_properties else None,
            'annotation_properties' if annotation_properties else None,
        ]).difference([None])
        for e in itertools.chain.from_iterable(
                getattr(owlready2.Ontology, c)(self) for c in categories):
            yield e
        if imported:
            for onto in self.get_imported_ontologies(recursive=True):
                for e in itertools.chain.from_iterable(
                        getattr(owlready2.Ontology, c)(onto)
                        for c in categories):
                    yield e

    def classes(self, imported=False):
        """Returns an generator over all classes.

        If `imported` is true, will imported classes are also returned.
        """
        return self.get_entities(
            imported=imported, classes=True, individuals=False,
            object_properties=False, data_properties=False,
            annotation_properties=False)

    def individuals(self, imported=False):
        """Returns an generator over all individuals.

        If `imported` is true, will imported individuals are also returned.
        """
        return self.get_entities(
            imported=imported, classes=False, individuals=True,
            object_properties=False, data_properties=False,
            annotation_properties=False)

    def object_properties(self, imported=False):
        """Returns an generator over all object properties.

        If `imported` is true, will imported object properties are also
        returned.
        """
        return self.get_entities(
            imported=imported, classes=False, individuals=False,
            object_properties=True, data_properties=False,
            annotation_properties=False)

    def data_properties(self, imported=False):
        """Returns an generator over all data properties.

        If `imported` is true, will imported data properties are also
        returned.
        """
        return self.get_entities(
            imported=imported, classes=False, individuals=False,
            object_properties=False, data_properties=True,
            annotation_properties=False)

    def annotation_properties(self, imported=False):
        """Returns a generator iterating over all annotation properties
        defined in the current ontology.

        If `imported` is true, annotation properties in imported ontologies
        will also be included.
        """
        return self.get_entities(
            imported=imported, classes=False, individuals=False,
            object_properties=False, data_properties=False,
            annotation_properties=True)

    def get_root_classes(self, imported=False):
        """Returns a list or root classes."""
        return [cls for cls in self.classes(imported=imported)
                if not cls.ancestors().difference(set([cls, owlready2.Thing]))]

    def get_root_object_properties(self, imported=False):
        """Returns a list of root object properties."""
        props = set(self.object_properties(imported=imported))
        return [p for p in props if not props.intersection(p.is_a)]

    def get_root_data_properties(self, imported=False):
        """Returns a list of root object properties."""
        props = set(self.data_properties(imported=imported))
        return [p for p in props if not props.intersection(p.is_a)]

    def get_roots(self, imported=False):
        """Returns all class, object_property and data_property roots."""
        roots = self.get_root_classes(imported=imported)
        roots.extend(self.get_root_object_properties(imported=imported))
        roots.extend(self.get_root_data_properties(imported=imported))
        return roots

    def get_by_label(self, label):
        """Returns entity by label.

        If several entities have the same label, only the one which is
        found first is returned.  A KeyError is raised if `label`
        cannot be found.
        """
        # Handle labels of the form 'namespace.label' recursively
        if '.' in label:
            head, sep, tail = label.partition('.')
            ns = self.get_namespace(head)
            return ns.ontology.get_by_label(tail)

        # Check for name in all categories in self
        for category in categories:
            method = getattr(self, category)
            for entity in method():
                if hasattr(entity, 'prefLabel') and label in entity.prefLabel:
                    return entity
                elif hasattr(entity, 'label') and label in entity.label:
                    return entity
                elif hasattr(entity, 'altLabel') and label in entity.altLabel:
                    return entity
        # Check for special names
        d = {
                'Nothing': owlready2.Nothing,
        }
        if label in d:
            return d[label]
        # Check whether `label` matches a Python class name of any category
        lst = [cls for cls in itertools.chain.from_iterable(
            getattr(self, category)() for category in categories)
             if hasattr(cls, '__name__') and cls.__name__ == label]
        if len(lst) == 1:
            return lst[0]
        elif len(lst) > 1:
            raise NoSuchLabelError('There is more than one Python class with '
                                   'name %r' % label)
        elif label is owlready2.Thing or label == 'Thing':
            return owlready2.Thing
        # Check imported ontologies
        for onto in self.imported_ontologies:
            onto.__class__ = self.__class__  # magically change type of onto
            try:
                return onto.get_by_label(label)
            except NoSuchLabelError:
                pass
        # Fallback to check whether we have a class in the current or any
        # of the imported ontologies whos name matches `label`
        for onto in [self] + self.imported_ontologies:
            lst = [cls for cls in onto.classes() if cls.__name__ == label]
            if len(lst) == 1:
                return lst[0]
            elif len(lst) > 1:
                raise NoSuchLabelError('There is more than one class with '
                                       'name %r' % label)
        # Label cannot be found
        raise NoSuchLabelError('Ontology "%s" has no such label: %s' % (
            self.name, label))

    def get_by_label_all(self, label):
        """Like get_by_label(), but returns a list of all entities with
        matching labels.
        """
        return [entity for entity in
                itertools.chain.from_iterable(
                    getattr(self, c)() for c in categories)
                if ((hasattr(entity, 'prefLabel') and label in entity.prefLabel) or
                    (hasattr(entity, 'label') and label in entity.label) or
                    (hasattr(entity, 'altLabel') and label in entity.altLabel))]

    def sync_reasoner(self, reasoner='HermiT', include_imported=False,
                      **kwargs):
        """Update current ontology by running the given reasoner.

        Supported values for `reasoner` are 'Pellet' and 'HermiT'.

        If `include_imported` is true, the reasoner will also reason
        over imported ontologies.  Note that this may be **very** slow
        with the current supported reasoners (FaCT++ seems must faster).

        Keyword arguments are passed to the underlying owlready2 function.
        """
        if reasoner == 'Pellet':
            sync = owlready2.sync_reasoner_pellet
        elif reasoner == 'HermiT':
            sync = owlready2.sync_reasoner_hermit
        else:
            raise ValueError('unknown reasoner %r.  Supported reasoners'
                             'are "Pellet" and "HermiT".', reasoner)

        if include_imported:
            with self:
                sync(**kwargs)
        else:
            sync([self], **kwargs)

    def sync_attributes(self, name_policy=None, name_prefix='',
                        sync_imported=False):
        """This method is intended to be called after you have added new
        classes (typically via Python) to make sure that attributes like
        `label` and `comments` are defined.

        If a class, object property, data property or annotation
        property in the current ontology has no label, the name of
        the corresponding Python class will be assigned as label.

        If a class, object property, data property or annotation
        property has no comment, it will be assigned the docstring of
        the corresponding Python class.

        `name_policy` specify wether and how the names in the ontology
        should be updated.  Valid values are:
          None          not changed
          "uuid"        `name_prefix` followed by a global unique id (UUID).
          "sequential"  `name_prefix` followed a sequantial number.
        EMMO conventions imply ``name_policy=='uuid'``.

        If `sync_imported` is true, all imported ontologies are also
        updated.
        """
        for cls in itertools.chain(
                self.classes(), self.object_properties(),
                self.data_properties(), self.annotation_properties()):
            if not hasattr(cls, 'prefLabel'):
                # no prefLabel - create new annotation property..
                with self:
                    class prefLabel(owlready2.label):
                        pass
                cls.prefLabel = [cls.__name__]
            elif not cls.prefLabel:
                cls.prefLabel.append(cls.__name__)
            if not cls.comment and hasattr(cls, '__doc__') and cls.__doc__:
                cls.comment.append(inspect.cleandoc(cls.__doc__))

        chain = itertools.chain(
            self.classes(), self.individuals(), self.object_properties(),
            self.data_properties(), self.annotation_properties())
        if name_policy == 'uuid':
            for obj in chain:
                obj.name = name_prefix + str(uuid.uuid4())
        elif name_policy == 'sequential':
            for obj in chain:
                n = 0
                while name_prefix + str(n) in self:
                    n += 1
                obj.name = name_prefix + str(n)
        elif name_policy is not None:
            raise TypeError('invalid name_policy: %r' % (name_policy, ))

        if sync_imported:
            for onto in self.imported_ontologies:
                onto.sync_attributes()

    def get_relations(self):
        """Returns a generator for all relations."""
        warnings.warn('Ontology.get_relations() is deprecated.  '
                      'Use onto.object_properties() instead.',
                      DeprecationWarning)
        return self.object_properties()

    def get_annotations(self, entity):
        """Returns a dict with annotations for `entity`.  Entity may be given
        either as a ThingClass object or as a label."""
        warnings.warn('Ontology.get_annotations(cls) is deprecated.  '
                      'Use cls.get_annotations() instead.', DeprecationWarning)

        if isinstance(entity, str):
            entity = self.get_by_label(entity)
        d = {'comment': getattr(entity, 'comment', '')}
        for a in self.annotation_properties():
            d[a.label.first()] = [
                o.strip('"') for s, p, o in
                self.get_triples(entity.storid, a.storid, None)]
        return d

    def get_branch(self, root, leafs=(), include_leafs=True,
                   strict_leafs=False, exclude=None, sort=False):
        """Returns a set with all direct and indirect subclasses of `root`.
        Any subclass found in the sequence `leafs` will be included in
        the returned list, but its subclasses will not.  The elements
        of `leafs` may be ThingClass objects or labels.

        Subclasses of any subclass found in the sequence `leafs` will
        be excluded from the returned list, where the elements of `leafs`
        may be ThingClass objects or labels.

        If `include_leafs` is true, the leafs are included in the returned
        list, otherwise they are not.

        If `strict_leafs` is true, any descendant of a leaf will be excluded
        in the returned set.

        If given, `exclude` may be a sequence of classes, including
        their subclasses, to exclude from the output.

        If `sort` is True, a list sorted according to depth and label
        will be returned instead of a set.
        """
        def _branch(root, leafs):
            if root not in leafs:
                branch = {root, }
                for c in root.subclasses():
                    # Defining a branch is actually quite tricky.  Consider
                    # the case:
                    #
                    #      L isA R
                    #      A isA L
                    #      A isA R
                    #
                    # where R is the root, L is a leaf and A is a direct
                    # child of both.  Logically, since A is a child of the
                    # leaf we want to skip A.  But a strait forward imple-
                    # mentation will see that A is a child of the root and
                    # include it.  Requireing that the R should be a strict
                    # parent of A solves this.
                    if root in c.get_parents(strict=True):
                        branch.update(_branch(c, leafs))
            else:
                branch = {root, } if include_leafs else set()
            return branch

        if isinstance(root, str):
            root = self.get_by_label(root)

        leafs = set(self.get_by_label(leaf) if isinstance(leaf, str)
                    else leaf for leaf in leafs)
        leafs.discard(root)

        if exclude:
            exclude = set(self.get_by_label(e) if isinstance(e, str)
                          else e for e in exclude)
            leafs.update(exclude)

        branch = _branch(root, leafs)

        # Exclude all descendants of any leaf
        if strict_leafs:
            descendants = root.descendants()
            for leaf in leafs:
                if leaf in descendants:
                    branch.difference_update(leaf.descendants(
                        include_self=False))

        if exclude:
            branch.difference_update(exclude)

        # Sort according to depth, then by label
        if sort:
            branch = sorted(sorted(branch, key=lambda x: asstring(x)),
                            key=lambda x: len(x.mro()))

        return branch

    def is_individual(self, entity):
        """Returns true if entity is an individual."""
        if isinstance(entity, str):
            entity = self.get_by_label(entity)
        return isinstance(entity, owlready2.Thing)

    # FIXME - deprecate this method as soon the ThingClass property
    #         `defined_class` works correct in Owlready2
    def is_defined(self, entity):
        """Returns true if the entity is a defined class."""
        if isinstance(entity, str):
            entity = self.get_by_label(entity)
        return hasattr(entity, 'equivalent_to') and bool(entity.equivalent_to)

    def get_version(self, as_iri=False):
        """Returns the version number of the ontology as inferred from the
        owl:versionIRI tag.

        If `as_iri` is True, the full versionIRI is returned.
        """
        versionIRI_storid = self.world._abbreviate(
            'http://www.w3.org/2002/07/owl#versionIRI')
        tokens = self.get_triples(s=self.storid, p=versionIRI_storid)
        if not tokens:
            raise TypeError('No versionIRI in Ontology %r' % self.base_iri)
        s, p, o = tokens[0]
        versionIRI = self.world._unabbreviate(o)
        if as_iri:
            return versionIRI
        else:
            return infer_version(self.base_iri, versionIRI)

    def set_version(self, version=None, version_iri=None):
        """Assign version to ontology by asigning owl:versionIRI.

        If `version` but not `version_iri` is provided, the version
        IRI will be the combination of `base_iri` and `version`.
        """
        versionIRI = 'http://www.w3.org/2002/07/owl#versionIRI'
        versionIRI_storid = self.world._abbreviate(versionIRI)
        if self._has_obj_triple_spo(s=self.storid, p=versionIRI_storid):
            self._del_obj_triple_spo(s=self.storid, p=versionIRI_storid)

        if not version_iri:
            if not version:
                raise TypeError(
                    'Either `version` or `version_iri` must be provided')
            version_iri = self.base_iri.rstrip('#/') + '/' + version

        self._add_obj_triple_spo(
            s=self.storid,
            p=self.world._abbreviate(versionIRI),
            o=self.world._abbreviate(version_iri),
        )

    def get_graph(self, **kwargs):
        """Returns a new graph object.  See  emmo.graph.OntoGraph.

        Note that this method requires the Python graphviz package.
        """
        from .graph import OntoGraph
        return OntoGraph(self, **kwargs)

    def common_ancestors(self, cls1, cls2):
        """Return a list of common ancestors for `cls1` and `cls2`."""
        return set(cls1.ancestors()).intersection(cls2.ancestors())

    def number_of_generations(self, descendant, ancestor):
        """ Return shortest distance from ancestor to descendant"""
        if ancestor not in descendant.ancestors():
            raise ValueError('Descendant is not a descendant of ancestor')
        return self._number_of_generations(descendant, ancestor, 0)

    def _number_of_generations(self, descendant, ancestor, n):
        """Recursive help function to number_of_generations(), return
        distance between a ancestor-descendant pair (n+1)."""
        if descendant.name == ancestor.name:
            return n
        return min(self._number_of_generations(parent, ancestor, n + 1)
                   for parent in descendant.get_parents()
                   if ancestor in parent.ancestors())

    def closest_common_ancestors(self, cls1, cls2):
        """Returns a list with closest_common_ancestor for cls1 and cls2"""
        distances = {}
        for ancestor in self.common_ancestors(cls1, cls2):
            distances[ancestor] = (self.number_of_generations(cls1, ancestor) +
                                   self.number_of_generations(cls2, ancestor))
        return [ancestor for ancestor, distance in distances.items()
                if distance == min(distances.values())]

    def closest_common_ancestor(self, *classes):
        """Returns closest_common_ancestor for the given classes."""
        mros = [cls.mro() for cls in classes]
        track = defaultdict(int)
        while mros:
            for mro in mros:
                cur = mro.pop(0)
                track[cur] += 1
                if track[cur] == len(classes):
                    return cur
                if len(mro) == 0:
                    mros.remove(mro)
        assert(0)  # should never be reached...



# -*- coding: utf-8 -*-
"""
A module for documenting ontologies.
"""
import os
import re
import time
import warnings
import shlex
import shutil
import subprocess
from textwrap import dedent
from tempfile import NamedTemporaryFile, TemporaryDirectory

import yaml
import owlready2

from .utils import asstring, camelsplit
from .graph import OntoGraph


class OntoDoc:
    """A class for helping documentating ontologies.

    Parameters
    ----------
    onto : Ontology instance
        The ontology that should be documented.
    style : dict | "html" | "markdown" | "markdown_tex"
        A dict defining the following template strings (and substitutions):

        :header: Formats an header.
            Substitutions: {level}, {label}
        :link: Formats a link.
           Substitutions: {name}
        :point: Formats a point (list item).
           Substitutions: {point}, {ontology}
        :points: Formats a list of points.  Used within annotations.
           Substitutions: {points}, {ontology}
        :annotation: Formats an annotation.
            Substitutions: {key}, {value}, {ontology}
        :substitutions: list of ``(regex, sub)`` pairs for substituting
            annotation values.
    """
    _markdown_style = dict(
        sep='\n',
        figwidth='{{ width={width:.0f}px }}',
        figure='![{caption}]({path}){figwidth}\n',
        header='\n{:#<{level}} {label}',
        link='[{name}]({lowerurl})',
        point='  - {point}\n',
        points='\n\n{points}\n',
        annotation='**{key}:** {value}\n',
        substitutions=[],
    )
    # Extra style settings for markdown+tex (e.g. pdf generation with pandoc)
    _markdown_tex_extra_style = dict(
        substitutions=[
            # logic/math symbols
            ('\u2200', r'$\\forall$'),
            ('\u2203', r'$\\exists$'),
            ('\u2206', r'$\\nabla$'),
            ('\u2227', r'$\\land$'),
            ('\u2228', r'$\\lor$'),
            ('\u2207', r'$\\nabla$'),
            ('\u2212', r'-'),
            ('->', r'$\\rightarrow$'),
            # uppercase greek letters
            ('\u0391', r'$\\Upalpha$'),
            ('\u0392', r'$\\Upbeta$'),
            ('\u0393', r'$\\Upgamma$'),
            ('\u0394', r'$\\Updelta$'),
            ('\u0395', r'$\\Upepsilon$'),
            ('\u0396', r'$\\Upzeta$'),
            ('\u0397', r'$\\Upeta$'),
            ('\u0398', r'$\\Uptheta$'),
            ('\u0399', r'$\\Upiota$'),
            ('\u039a', r'$\\Upkappa$'),
            ('\u039b', r'$\\Uplambda$'),
            ('\u039c', r'$\\Upmu$'),
            ('\u039d', r'$\\Upnu$'),
            ('\u039e', r'$\\Upxi$'),
            ('\u039f', r'$\\Upomekron$'),
            ('\u03a0', r'$\\Uppi$'),
            ('\u03a1', r'$\\Uprho$'),
            ('\u03a3', r'$\\Upsigma$'),  # no \u0302
            ('\u03a4', r'$\\Uptau$'),
            ('\u03a5', r'$\\Upupsilon$'),
            ('\u03a6', r'$\\Upvarphi$'),
            ('\u03a7', r'$\\Upchi$'),
            ('\u03a8', r'$\\Uppsi$'),
            ('\u03a9', r'$\\Upomega$'),
            # lowercase greek letters
            ('\u03b1', r'$\\upalpha$'),
            ('\u03b2', r'$\\upbeta$'),
            ('\u03b3', r'$\\upgamma$'),
            ('\u03b4', r'$\\updelta$'),
            ('\u03b5', r'$\\upepsilon$'),
            ('\u03b6', r'$\\upzeta$'),
            ('\u03b7', r'$\\upeta$'),
            ('\u03b8', r'$\\uptheta$'),
            ('\u03b9', r'$\\upiota$'),
            ('\u03ba', r'$\\upkappa$'),
            ('\u03bb', r'$\\uplambda$'),
            ('\u03bc', r'$\\upmu$'),
            ('\u03bd', r'$\\upnu$'),
            ('\u03be', r'$\\upxi$'),
            ('\u03bf', r'o'),  # no \upomicron
            ('\u03c0', r'$\\uppi$'),
            ('\u03c1', r'$\\uprho$'),
            ('\u03c2', r'$\\upvarsigma$'),
            ('\u03c3', r'$\\upsigma$'),
            ('\u03c4', r'$\\uptau$'),
            ('\u03c5', r'$\\upupsilon$'),
            ('\u03c6', r'$\\upvarphi$'),
            ('\u03c7', r'$\\upchi$'),
            ('\u03c8', r'$\\uppsi$'),
            ('\u03c9', r'$\\upomega$'),
            # acutes, accents, etc...
            ('\u03ae', r"$\\acute{\\upeta}$"),
            ('\u1e17', r"$\\acute{\\bar{\\mathrm{e}}}$"),
            ('\u03ac', r"$\\acute{\\upalpha}$"),
            ('\u00e1', r"$\\acute{\\mathrm{a}}$"),
            ('\u03cc', r"$\\acute{o}$"),  # no \upomicron
            ('\u014d', r"$\\bar{\\mathrm{o}}$"),
            ('\u1f45', r'$\\acute{o}$'),  # no \omicron

        ],
    )
    _html_style = dict(
        sep='<p>\n',
        figwidth='width="{width:.0f}"',
        figure='<img src="{path}" alt="{caption}"{figwidth}>',
        header='<h{level} id="{lowerlabel}">{label}</h{level}>',
        link='<a href="{lowerurl}">{name}</a>',
        point='      <li>{point}</li>\n',
        points='    <ul>\n      {points}\n    </ul>\n',
        annotation='  <dd><strong>{key}:</strong>\n{value}  </dd>\n',
        substitutions=[
            (r'\n\n', r'<p>'),
            (r'\n', r'<br>\n'),
            (r'&', r"&#8210;"),
            (r'<p>', r'<p>\n\n'),
            (r'\u2018([^\u2019]*)\u2019', r'<q>\1</q>'),
            (r'\u2019', r"'"),
            (r'\u2260', r"&ne;"),
            (r'\u2264', r"&le;"),
            (r'\u2265', r"&ge;"),
            (r'\u226A', r"&x226A;"),
            (r'\u226B', r"&x226B;"),
            (r'"Y$', r""),  # strange noice added by owlready2
        ],
    )

    def __init__(self, onto, style='markdown'):
        if isinstance(style, str):
            if style == 'markdown_tex':
                style = self._markdown_style.copy()
                style.update(self._markdown_tex_extra_style)
            else:
                style = getattr(self, '_%s_style' % style)
        self.onto = onto
        self.style = style
        self.url_regex = re.compile(r'https?:\/\/[^\s ]+')

    def get_default_template(self):
        """Returns default template."""
        title = os.path.splitext(
            os.path.basename(self.onto.base_iri.rstrip('/#')))[0]
        irilink = self.style.get('link', '{name}').format(
            name=self.onto.base_iri, url=self.onto.base_iri,
            lowerurl=self.onto.base_iri)
        s = dedent('''\
        %HEADER {title}
        Documentation of {irilink}

        %HEADER Relations level=2
        %ALL object_properties

        %HEADER Classes level=2
        %ALL classes

        %HEADER Individuals level=2
        %ALL individuals

        %HEADER Appendix               level=1
        %HEADER "Relation taxonomies"  level=2
        %ALLFIG object_properties

        %HEADER "Class taxonomies"     level=2
        %ALLFIG classes
        ''').format(ontology=self.onto, title=title, irilink=irilink)
        return s

    def get_header(self, label, header_level=1):
        """Returns `label` formatted as a header of given level."""
        header_style = self.style.get('header', '{label}\n')
        return header_style.format(
            '', level=header_level, label=label, lowerlabel=label.lower())

    def get_figure(self, path, caption='', width=None):
        """Returns a formatted insert-figure-directive."""
        figwidth_style = self.style.get('figwidth', '')
        figure_style = self.style.get('figure', '')
        figwidth = figwidth_style.format(width=width) if width else ''
        return figure_style.format(path=path, caption=caption,
                                   figwidth=figwidth)

    def itemdoc(self, item, header_level=3, show_disjoints=False):
        """Returns documentation of `item`.

        Parameters
        ----------
        item : obj | label
            The class, individual or relation to document.
        header_level : int
            Header level. Defaults to 3.
        show_disjoints : Bool
            Whether to show `disjoint_with` relations.
        """
        onto = self.onto
        if isinstance(item, str):
            item = self.onto.get_by_label(item)

        header_style = self.style.get('header', '{label}\n')
        link_style = self.style.get('link', '{name}')
        point_style = self.style.get('point', '{point}')
        points_style = self.style.get('points', '{points}')
        annotation_style = self.style.get('annotation', '{key}: {value}\n')
        substitutions = self.style.get('substitutions', [])

        # Logical "sorting" of annotations
        order = dict(definition='00', axiom='01', theorem='02',
                     elucidation='03', domain='04', range='05', example='06')

        doc = []

        # Header
        label = item.prefLabel.first()
        doc.append(header_style.format(
            '', level=header_level, label=label, lowerlabel=label.lower()))

        # Add iri
        doc.append(annotation_style.format(
            key='IRI', value=asstring(item.iri, link_style), ontology=onto))

        # Add annotations
        if isinstance(item, owlready2.Thing):
            annotations = item.get_individual_annotations()
        else:
            annotations = item.get_annotations()

        for key in sorted(annotations.keys(),
                          key=lambda key: order.get(key, key)):
            for value in annotations[key]:
                if self.url_regex.match(value):
                    doc.append(annotation_style.format(
                        key=key.capitalize(),
                        value=asstring(value, link_style)))
                else:
                    for reg, sub in substitutions:
                        value = re.sub(reg, sub, value)
                    doc.append(annotation_style.format(
                        key=key.capitalize(), value=value))

        # ...add relations from is_a
        points = []
        nonProp = (owlready2.ThingClass,  # owlready2.Restriction,
                   owlready2.And, owlready2.Or, owlready2.Not)
        for p in item.is_a:
            if (isinstance(p, nonProp) or
                (isinstance(item, owlready2.PropertyClass) and
                 isinstance(p, owlready2.PropertyClass))):
                points.append(point_style.format(
                    point='is_a ' + asstring(p, link_style), ontology=onto))
            else:
                points.append(point_style.format(
                    point=asstring(p, link_style), ontology=onto))

        # ...add equivalent_to relations
        for e in item.equivalent_to:
            points.append(point_style.format(
                point='equivalent_to ' + asstring(e, link_style)))

        # ...add disjoint_with relations
        if show_disjoints and hasattr(item, 'disjoint_with'):
            s = set(item.disjoint_with(reduce=True))
            points.append(point_style.format(
                point='disjoint_with ' + ', '.join(asstring(e, link_style)
                                                   for e in s), ontology=onto))

        # ...add disjoint_unions
        if hasattr(item, 'disjoint_unions'):
            for u in item.disjoint_unions:
                s = ', '.join(asstring(e, link_style) for e in u)
                points.append(point_style.format(
                    point='disjoint_union_of ' + s, ontology=onto))

        # ...add inverse_of relations
        if hasattr(item, 'inverse_property') and item.inverse_property:
            points.append(point_style.format(
                point='inverse_of ' + asstring(
                    item.inverse_property, link_style)))

        # ...add domain restrictions
        for d in getattr(item, 'domain', ()):
            points.append(point_style.format(
                point='domain ' + asstring(d, link_style)))

        # ...add range restrictions
        for d in getattr(item, 'range', ()):
            points.append(point_style.format(
                point='range ' + asstring(d, link_style)))

        # Add relations
        if points:
            value = points_style.format(
                points=''.join(points), ontology=onto)
            doc.append(annotation_style.format(
                key='Relations', value=value, ontology=onto))

        # Instances (individuals)
        if hasattr(item, 'instances'):
            points = []
            for e in [i for i in item.instances() if item in i.is_instance_of]:
                points.append(point_style.format(
                    point=asstring(e, link_style), ontology=onto))
            if points:
                value = points_style.format(
                    points=''.join(points), ontology=onto)
                doc.append(annotation_style.format(
                    key='Individuals', value=value, ontology=onto))

        return '\n'.join(doc)

    def itemsdoc(self, items, header_level=3):
        """Returns documentation of `items`."""
        sep_style = self.style.get('sep', '\n')
        doc = []
        for item in items:
            doc.append(self.itemdoc(item, header_level))
            doc.append(sep_style.format(ontology=self.onto))
        return '\n'.join(doc)


class attrdict(dict):
    """A dict with attribute access.

    Note that methods like key() and update() may be overridden."""
    def __init__(self, *args, **kwargs):
        super(attrdict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class InvalidTemplateError(NameError):
    """Raised on errors in template files."""


def get_options(opts, **kw):
    """Returns a dict with options from the sequence `opts` with
    "name=value" pairs. Valid option names and default values are
    provided with the keyword arguments."""
    d = attrdict(kw)
    for opt in opts:
        if '=' not in opt:
            raise InvalidTemplateError('Missing "=" in template option: %r' %
                                       opt)
        name, value = opt.split('=', 1)
        if name not in d:
            raise InvalidTemplateError('Invalid template option: %r' % name)
        t = type(d[name])
        d[name] = t(value)
    return d


class DocPP:
    """Documentation pre-processor.

    It supports the following features:

      * Comment lines

            %% Comment line...

      * Insert header with given level

            %HEADER label [level=1]

      * Insert figure with optional caption and width. `filepath`
        should be relative to `basedir`.  If width is 0, no width will
        be specified.

            %FIGURE filepath [caption='' width=0px]

      * Include other markdown files.  Header levels may be up or down with
        `shift`

            %INCLUDE filepath [shift=0]

      * Insert generated documentation for ontology entity.  The header
        level may be set with `header_level`.

            %ENTITY name [header_level=3]

      * Insert generated documentation for ontology branch `name`.  Options:
          - header_level: Header level.
          - terminated: Whether to branch should be terminated at all branch
            names in the final document.
          - include_leafs: Whether to include leaf.

            %BRANCH name [header_level=3 terminated=1 include_leafs=0]

      * Insert generated figure of ontology branch `name`.  The figure
        is written to `path`.  The default path is `figdir`/`name`,
        where `figdir` is given at class initiation. It is recommended
        to exclude the file extension from `path`.  In this case, the
        default figformat will be used (and easily adjusted to the
        correct format required by the backend). `leafs` may be a comma-
        separated list of leaf node names.

            %BRANCHFIG name [path='' caption='' terminated=1 include_leafs=1
                             strict_leafs=1, width=0px leafs='' relations=all
                             edgelabels=0]

      * This is a combination of the %HEADER and %BRANCHFIG directives.

            %BRANCHHEAD name [level=2  path='' caption='' terminated=1
                              include_leafs=1 width=0px leafs='']

      * This is a combination of the %HEADER, %BRANCHFIG and %BRANCH
        directives. It inserts documentation of branch `name`, with a
        header followed by a figure and then documentation of each
        element.

            %BRANCHDOC name [level=2  path='' caption='' terminated=1
                             width=0px leafs='']

      * Insert generated documentation for all entities of the given type.
        Valid values of `type` are: "classes", "individuals",
        "object_properties", "data_properties", "annotations_properties"

            %ALL type [header_level=3]

      * Insert generated figure of all entities of the given type.
        Valid values of `type` are: "classes", "object_properties" and
        "data_properties".

            %ALLFIG type

    Parameters
    ----------
    template : str
        Input template.
    ontodoc : OntoDoc instance
        Instance of OntoDoc
    basedir : str
        Base directory for including relative file paths.
    figdir : str
        Default directory to store generated figures.
    figformat : str
        Default format for generated figures.
    figscale : float
        Default scaling of generated figures.
    maxwidth : float
        Maximum figure width.  Figures larger than this will be rescaled.
    imported : bool
        Whether to include imported entities.
    """

    # FIXME - this class should be refractured:
    #   * Instead of rescan the entire document for each pre-processer
    #     directive, we should scan the source like by line and handle
    #     each directive as they occour.  The current implementation has
    #     a lot of dublicated code.
    #   * Instead of modifying the source in-place, we should copy to a
    #     result list. This will make good error reporting much easier.
    #   * Branch leaves are only looked up in the file witht the %BRANCH
    #     directive, not in all included files as expedted.

    def __init__(self, template, ontodoc, basedir='.', figdir='genfigs',
                 figformat='png', figscale=1.0, maxwidth=None,
                 imported=False):
        self.lines = template.split('\n')
        self.ontodoc = ontodoc
        self.basedir = basedir
        self.figdir = os.path.join(basedir, figdir)
        self.figformat = figformat
        self.figscale = figscale
        self.maxwidth = maxwidth
        self.imported = imported
        self._branch_cache = None
        self._processed = False  # Whether process() has been called

    def __str__(self):
        return self.get_buffer()

    def get_buffer(self):
        """Returns the current buffer."""
        return '\n'.join(self.lines)

    def copy(self):
        """Returns a copy of self."""
        docpp = DocPP('', self.ontodoc, self.basedir, figformat=self.figformat,
                      figscale=self.figscale, maxwidth=self.maxwidth)
        docpp.lines[:] = self.lines
        docpp.figdir = self.figdir
        return docpp

    def get_branches(self):
        """Returns a list with all branch names as specified with %BRANCH
        (in current and all included documents).  The returned value is
        cached for efficiency purposes and so that it is not lost after
        processing branches."""
        if self._branch_cache is None:
            names = []
            docpp = self.copy()
            docpp.process_includes()
            for line in docpp.lines:
                if line.startswith('%BRANCH'):
                    names.append(shlex.split(line)[1])
            self._branch_cache = names
        return self._branch_cache

    def shift_header_levels(self, shift):
        """Shift header level of all hashtag-headers in buffer.  Underline
        headers are ignored."""
        if not shift:
            return
        pat = re.compile('^#+ ')
        for i, line in enumerate(self.lines):
            m = pat.match(line)
            if m:
                if shift > 0:
                    self.lines[i] = '#' * shift + line
                elif shift < 0:
                    n = m.end()
                    if shift > n:
                        self.lines[i] = line.lstrip('# ')
                    else:
                        self.lines[i] = line[n:]

    def process_comments(self):
        """Strips out comment lines starting with "%%"."""
        self.lines = [line for line in self.lines if not line.startswith('%%')]

    def process_headers(self):
        """Expand all %HEADER specifications."""
        for i, line in reversed(list(enumerate(self.lines))):
            if line.startswith('%HEADER '):
                tokens = shlex.split(line)
                name = tokens[1]
                opts = get_options(tokens[2:], level=1)
                del self.lines[i]
                self.lines[i: i] = self.ontodoc.get_header(
                    name, int(opts.level)).split('\n')

    def process_figures(self):
        """Expand all %FIGURE specifications."""
        for i, line in reversed(list(enumerate(self.lines))):
            if line.startswith('%FIGURE '):
                tokens = shlex.split(line)
                path = tokens[1]
                opts = get_options(tokens[2:], caption='', width=0)
                del self.lines[i]
                self.lines[i: i] = self.ontodoc.get_figure(
                    os.path.join(self.basedir, path),
                    caption=opts.caption, width=opts.width).split('\n')

    def process_entities(self):
        """Expand all %ENTITY specifications."""
        for i, line in reversed(list(enumerate(self.lines))):
            if line.startswith('%ENTITY '):
                tokens = shlex.split(line)
                name = tokens[1]
                opts = get_options(tokens[2:], header_level=3)
                del self.lines[i]
                self.lines[i: i] = self.ontodoc.itemdoc(
                    name, int(opts.header_level)).split('\n')

    def process_branches(self):
        """Expand all %BRANCH specifications."""
        # Get all branch names in final document
        names = self.get_branches()
        for i, line in reversed(list(enumerate(self.lines))):
            if line.startswith('%BRANCH '):
                tokens = shlex.split(line)
                name = tokens[1]
                opts = get_options(tokens[2:], header_level=3, terminated=1,
                                   include_leafs=0)
                leafs = names if opts.terminated else ()
                branch = self.ontodoc.onto.get_branch(name, leafs,
                                                      opts.include_leafs)
                del self.lines[i]
                self.lines[i: i] = self.ontodoc.itemsdoc(
                    branch, int(opts.header_level)).split('\n')

    def _make_branchfig(self, name, path, terminated, include_leafs,
                        strict_leafs, width, leafs, relations, edgelabels,
                        rankdir, legend):
        """Help method for process_branchfig().

        Args:
            name: name of branch root
            path: optional figure path name
            include_leafs: whether to include leafs
            strict_leafs: whether strictly exclude leafs descendants
            terminated: whether the graph should be terminated at leaf nodes
            width: optional figure width
            leafs: optional leafs node names for graph termination
            relations: comma-separated list of relations to include
            edgelabels: whether to include edgelabels
            rankdir: graph direction (BT, TB, RL, LR)
            legend: whether to add legend

        Returns:
            filepath: path to generated figure
            leafs: used list of leaf node names
            width: actual figure width
        """
        onto = self.ontodoc.onto
        if leafs:
            if isinstance(leafs, str):
                leafs = leafs.split(',')
        elif terminated:
            leafs = set(self.get_branches())
            leafs.discard(name)
        else:
            leafs = None
        if path:
            figdir = os.path.dirname(path)
            formatext = os.path.splitext(path)[1]
            if formatext:
                format = formatext.lstrip('.')
            else:
                format = self.figformat
                path += '.' + format
        else:
            figdir = self.figdir
            format = self.figformat
            term = 'T' if terminated else ''
            path = os.path.join(figdir, name + term) + '.' + format

        # Create graph
        graph = OntoGraph(onto, graph_attr={'rankdir': rankdir})
        graph.add_branch(root=name, leafs=leafs, include_leafs=include_leafs,
                         strict_leafs=strict_leafs, relations=relations,
                         edgelabels=edgelabels)
        if legend:
            graph.add_legend()

        if not width:
            figwidth, figheight = graph.get_figsize()
            width = self.figscale * figwidth
            if self.maxwidth and width > self.maxwidth:
                width = self.maxwidth

        filepath = os.path.join(self.basedir, path)
        destdir = os.path.dirname(filepath)
        if not os.path.exists(destdir):
            os.makedirs(destdir)
        graph.save(filepath, format=format)
        return filepath, leafs, width

    def process_branchfigs(self):
        """Process all %BRANCHFIG directives."""
        for i, line in reversed(list(enumerate(self.lines))):
            if line.startswith('%BRANCHFIG '):
                tokens = shlex.split(line)
                name = tokens[1]
                opts = get_options(
                    tokens[2:], path='', caption='', terminated=1,
                    include_leafs=1, strict_leafs=1, width=0, leafs='',
                    relations='all', edgelabels=0, rankdir='BT', legend=1)
                filepath, leafs, width = self._make_branchfig(
                    name, opts.path, opts.terminated, opts.include_leafs,
                    opts.strict_leafs, opts.width, opts.leafs, opts.relations,
                    opts.edgelabels, opts.rankdir, opts.legend)

                del self.lines[i]
                self.lines[i: i] = self.ontodoc.get_figure(
                    filepath, caption=opts.caption, width=width).split('\n')

    def process_branchdocs(self):
        """Process all %BRANCHDOC and  %BRANCHEAD directives."""
        onto = self.ontodoc.onto
        for i, line in reversed(list(enumerate(self.lines))):
            if (line.startswith('%BRANCHDOC ') or
                    line.startswith('%BRANCHHEAD ')):
                with_branch = True if line.startswith('%BRANCHDOC ') else False
                tokens = shlex.split(line)
                name = tokens[1]
                title = camelsplit(name)
                title = title[0].upper() + title[1:] + ' branch'
                opts = get_options(tokens[2:], level=2, path='', title=title,
                                   caption=title + '.', terminated=1,
                                   strict_leafs=1, width=0,
                                   leafs='', relations='all', edgelabels=0,
                                   rankdir='BT', legend=1)

                include_leafs = 1
                filepath, leafs, width = self._make_branchfig(
                    name, opts.path, opts.terminated, include_leafs,
                    opts.strict_leafs, opts.width, opts.leafs, opts.relations,
                    opts.edgelabels, opts.rankdir, opts.legend)

                sec = []
                sec.append(
                    self.ontodoc.get_header(opts.title, int(opts.level)))
                sec.append(
                    self.ontodoc.get_figure(filepath, caption=opts.caption,
                                            width=width))
                if with_branch:
                    include_leafs = 0
                    branch = onto.get_branch(name, leafs, include_leafs)
                    sec.append(
                        self.ontodoc.itemsdoc(branch, int(opts.level + 1)))

                del self.lines[i]
                self.lines[i: i] = sec

    def process_alls(self):
        """Expand all %ALL specifications."""
        onto = self.ontodoc.onto
        for i, line in reversed(list(enumerate(self.lines))):
            if line.startswith('%ALL '):
                tokens = shlex.split(line)
                type = tokens[1]
                opts = get_options(tokens[2:], header_level=3)
                if type == 'classes':
                    items = onto.classes(imported=self.imported)
                elif type in ('object_properties', 'relations'):
                    items = onto.object_properties(imported=self.imported)
                elif type == 'data_properties':
                    items = onto.data_properties(imported=self.imported)
                elif type == 'annotation_properties':
                    items = onto.annotation_properties(imported=self.imported)
                elif type == 'individuals':
                    items = onto.individuals(imported=self.imported)
                else:
                    raise InvalidTemplateError(
                        'Invalid argument to %%ALL: %s' % type)
                items = sorted(items, key=lambda x: asstring(x))
                del self.lines[i]
                self.lines[i: i] = self.ontodoc.itemsdoc(
                    items, int(opts.header_level)).split('\n')

    def process_allfig(self):
        """Process all %ALLFIG directives."""
        onto = self.ontodoc.onto
        for i, line in reversed(list(enumerate(self.lines))):
            if line.startswith('%ALLFIG '):
                tokens = shlex.split(line)
                type = tokens[1]
                opts = get_options(tokens[2:], path='', level=3, terminated=0,
                                   include_leafs=1, strict_leafs=1, width=0,
                                   leafs='', relations='isA', edgelabels=0,
                                   rankdir='BT', legend=1)
                if type == 'classes':
                    roots = onto.get_root_classes(imported=self.imported)
                elif type in ('object_properties', 'relations'):
                    roots = onto.get_root_object_properties(imported=self.imported)
                elif type == 'data_properties':
                    roots = onto.get_root_data_properties(imported=self.imported)
                else:
                    raise InvalidTemplateError(
                        'Invalid argument to %%ALL: %s' % type)

                sec = []
                for root in roots:
                    name = asstring(root)
                    filepath, leafs, width = self._make_branchfig(
                        name, opts.path, opts.terminated, opts.include_leafs,
                        opts.strict_leafs, opts.width, opts.leafs,
                        opts.relations, opts.edgelabels, opts.rankdir,
                        opts.legend)
                    title = 'Taxonomy of %s.' % name
                    sec.append(
                        self.ontodoc.get_header(title, int(opts.level)))
                    caption = 'Taxonomy of %s.' % name
                    sec.extend(self.ontodoc.get_figure(
                        filepath, caption=caption, width=width).split('\n'))

                del self.lines[i]
                self.lines[i: i] = sec

    def process_includes(self):
        """Process all %INCLUDE directives."""
        for i, line in reversed(list(enumerate(self.lines))):
            if line.startswith('%INCLUDE '):
                tokens = shlex.split(line)
                filepath = tokens[1]
                opts = get_options(tokens[2:], shift=0)
                with open(os.path.join(self.basedir, filepath), 'rt') as f:
                    docpp = DocPP(
                        f.read(), self.ontodoc,
                        basedir=os.path.dirname(filepath),
                        figformat=self.figformat, figscale=self.figscale,
                        maxwidth=self.maxwidth)
                    docpp.figdir = self.figdir
                if opts.shift:
                    docpp.shift_header_levels(int(opts.shift))
                docpp.process()
                del self.lines[i]
                self.lines[i: i] = docpp.lines

    def process(self):
        """Perform all pre-processing steps."""
        if not self._processed:
            self.process_comments()
            self.process_headers()
            self.process_figures()
            self.process_entities()
            self.process_branches()
            self.process_branchfigs()
            self.process_branchdocs()
            self.process_alls()
            self.process_allfig()
            self.process_includes()
            self._processed = True

    def write(self, outfile, format=None, pandoc_option_files=(),
              pandoc_options=(), genfile=None, verbose=True):
        """Writes documentation to `outfile`.

        Parameters
        ----------
        outfile : str
            File that the documentation is written to.
        format : str
            Output format.  If it is "md" or "simple-html",
            the built-in template generator is used.  Otherwise
            pandoc is used.  If not given, the format is inferred
            from the `outfile` name extension.
        pandoc_option_files : sequence
            Sequence with command line arguments provided to pandoc.
        pandoc_options : sequence
            Additional pandoc options overriding options read from
        `pandoc_option_files`.
        genfile : str
            Store temporary generated markdown input file to pandoc
            to this file (for debugging).
        verbose : bool
            Whether to show some messages when running pandoc.
        """
        self.process()
        content = self.get_buffer()

        substitutions = self.ontodoc.style.get('substitutions', [])
        for reg, sub in substitutions:
            content = re.sub(reg, sub, content)

        format = get_format(outfile, format)
        if format not in ('simple-html', 'markdown', 'md'):  # Run pandoc
            if not genfile:
                f = NamedTemporaryFile(mode='w+t', suffix='.md')
                f.write(content)
                f.flush()
                genfile = f.name
            else:
                with open(genfile, 'wt') as f:
                    f.write(content)
            run_pandoc(genfile, outfile, format,
                       pandoc_option_files=pandoc_option_files,
                       pandoc_options=pandoc_options,
                       verbose=verbose)
        else:
            if verbose:
                print('Writing:', outfile)
            with open(outfile, 'wt') as f:
                f.write(content)


def load_pandoc_option_file(yamlfile):
    """Loads pandoc options from `yamlfile` and return a list with
    corresponding pandoc command line arguments."""
    with open(yamlfile) as f:
        d = yaml.safe_load(f)
    options = d.pop('input-files', [])
    variables = d.pop('variables', {})

    for k, v in d.items():
        if isinstance(v, bool):
            if v:
                options.append('--%s' % k)
        else:
            options.append('--%s=%s' % (k, v))

    for k, v in variables.items():
        if k == 'date' and v == 'now':
            v = time.strftime('%B %d, %Y')
        options.append('--variable=%s:%s' % (k, v))

    return options


def append_pandoc_options(options, updates):
    """Append `updates` to pandoc options `options`.

    Parameters
    ----------
    options : sequence
        Sequence with initial Pandoc options.
    updates : sequence of str
        Sequence of strings of the form "--longoption=value", where
        ``longoption`` is a valid pandoc long option and ``value`` is the
        new value.  The "=value" part is optional.

        Strings of the form "no-longoption" will filter out "--longoption"
        from `options`.

    Returns
    -------
    new_options : list
        Updated pandoc options.
    """
    # Valid pandoc options starting with "--no-XXX"
    no_options = set('no-highlight')

    if not updates:
        return list(options)

    u = {}
    for s in updates:
        k, sep, v = s.partition('=')
        u[k.lstrip('-')] = v if sep else None
        filter_out = set(k for k, v in u.items()
                         if k.startswith('no-') and k not in no_options)
        _filter_out = set('--' + k[3:] for k in filter_out)
        new_options = [opt for opt in options
                       if opt.partition('=')[0] not in _filter_out]
        new_options.extend(['--%s' % k if v is None else '--%s=%s' % (k, v)
                            for k, v in u.items()
                            if k not in filter_out])
    return new_options


def run_pandoc(genfile, outfile, format, pandoc_option_files=(),
               pandoc_options=(), verbose=True):
    """Runs pandoc.

    Parameters
    ----------
    genfile : str
        Name of markdown input file.
    outfile : str
        Output file name.
    format : str
        Output format.
    pandoc_option_files : sequence
        List of files with additional pandoc options.  Default is to read
        "pandoc-options.yaml" and "pandoc-FORMAT-options.yml", where
        `FORMAT` is the output format.
    pandoc_options : sequence
        Additional pandoc options overriding options read from
        `pandoc_option_files`.
    verbose : bool
        Whether to print the pandoc command before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the pandoc process returns with non-zero status.  The `returncode`
        attribute will hold the exit code.
    """
    # Create pandoc argument list
    args = [genfile]
    files = ['pandoc-options.yaml', 'pandoc-%s-options.yaml' % format]
    if pandoc_option_files:
        files = pandoc_option_files
    for fname in files:
        if os.path.exists(fname):
            args.extend(load_pandoc_option_file(fname))
        else:
            warnings.warn('missing pandoc option file: %s' % fname)

    # Update pandoc argument list
    args = append_pandoc_options(args, pandoc_options)

    # pdf output requires a special attention...
    if format == 'pdf':
        pdf_engine = 'pdflatex'
        for arg in args:
            if arg.startswith('--pdf-engine'):
                pdf_engine = arg.split('=', 1)[1]
                break
        with TemporaryDirectory() as tmpdir:
            run_pandoc_pdf(tmpdir, pdf_engine, outfile, args, verbose=verbose)
    else:
        args.append('--output=%s' % outfile)
        cmd = ['pandoc'] + args
        if verbose:
            print()
            print('* Executing command:')
            print(' '.join(shlex.quote(s) for s in cmd))
        subprocess.check_call(cmd)


def run_pandoc_pdf(latex_dir, pdf_engine, outfile, args, verbose=True):
    """Run pandoc for pdf generation."""
    basename = os.path.join(latex_dir, os.path.splitext(
        os.path.basename(outfile))[0])

    # Run pandoc
    texfile = basename + '.tex'
    args.append('--output=%s' % texfile)
    cmd = ['pandoc'] + args
    if verbose:
        print()
        print('* Executing commands:')
        print(' '.join(shlex.quote(s) for s in cmd))
    subprocess.check_call(cmd)

    # Fixing tex output
    texfile2 = basename + '2.tex'
    with open(texfile, 'rt') as f:
        content = f.read().replace(r'\$\Uptheta\$', r'$\Uptheta$')
    with open(texfile2, 'wt') as f:
        f.write(content)

    # Run latex
    pdffile = basename + '2.pdf'
    cmd = [pdf_engine, texfile2, '-halt-on-error',
           '-output-directory=%s' % latex_dir]
    if verbose:
        print()
        print(' '.join(shlex.quote(s) for s in cmd))
    output = subprocess.check_output(cmd, timeout=60)
    output = subprocess.check_output(cmd, timeout=60)

    # Workaround for non-working "-output-directory" latex option
    if not os.path.exists(pdffile):
        if os.path.exists(os.path.basename(pdffile)):
            pdffile = os.path.basename(pdffile)
            for ext in 'aux', 'out', 'toc', 'log':
                filename = os.path.splitext(pdffile)[0] + '.' + ext
                if os.path.exists(filename):
                    os.remove(filename)
        else:
            print()
            print(output)
            print()
            raise RuntimeError('latex did not produce pdf file: ' + pdffile)

    # Copy pdffile
    if not os.path.exists(outfile) or not os.path.samefile(pdffile, outfile):
        if verbose:
            print()
            print('move %s to %s' % (pdffile, outfile))
        shutil.move(pdffile, outfile)


def get_format(outfile, format=None):
    """Infer format from outfile and format."""
    if format is None:
        format = os.path.splitext(outfile)[1]
    if not format:
        format = 'html'
    if format.startswith('.'):
        format = format[1:]
    return format


def get_style(format):
    """Infer style from output format."""
    if format == 'simple-html':
        style = 'html'
    elif format in ('tex', 'latex', 'pdf'):
        style = 'markdown_tex'
    else:
        style = 'markdown'
    return style


def get_figformat(format):
    """Infer preferred figure format from output format."""
    if format == 'pdf':
        figformat = 'pdf'  # XXX
    elif 'html' in format:
        figformat = 'svg'
    else:
        figformat = 'png'
    return figformat


def get_maxwidth(format):
    """Infer preferred max figure width from output format."""
    if format == 'pdf':
        maxwidth = 668
    else:
        maxwidth = 1024
    return maxwidth


def get_docpp(ontodoc, infile, figdir='genfigs', figformat='png',
              maxwidth=None):
    """Read `infile` and return a new docpp instance."""
    if infile:
        with open(infile, 'rt') as f:
            template = f.read()
        basedir = os.path.dirname(infile)
    else:
        template = ontodoc.get_default_template()
        basedir = '.'

    docpp = DocPP(template, ontodoc, basedir=basedir, figdir=figdir,
                  figformat=figformat, maxwidth=maxwidth)

    return docpp



#!/usr/bin/env python3
import sys
import os

# Add emmo to sys path
thisdir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(1, os.path.abspath(os.path.join(thisdir, '..', '..')))
from emmo.nadict import NADict  # noqa: E402, F401


n = NADict(a=1, b=NADict(c=3, d=4))

assert n.a == 1
assert n.b.c == 3
assert n.b.d == 4
assert n['b.c'] == 3



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Step 1 - Generate metadata from ontology
----------------------------------------
This step takes the ontology from the vertical case and generates
metadata from it.

Here we use DLite as metadata framework, which is a SOFT
implementation in C.  Other frameworks, like CUDS, could equally well
have been used, but the metadata structure would have been different.

This generator maps OWL classes to dlite metadata entities which are
placed into a collection.  OWL class properties are mapped into
relations between the entities in the collection. The only relations
that are treated especially are `has_property` and `is_property_for`,
that are mapped into properties of the generated metadata entities.

The generated  metadata is finally serialised into a JSON file.
"""
import os

from emmo import get_ontology
from emmo2meta import EMMO2Meta


# Load our ontology from the vertical case
ontopath = os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..', 'vertical', 'usercase_ontology.owl'))
onto = get_ontology(ontopath)
onto.load()


# hmm, the base iri has cnanged... - bug in owlready2?
onto.base_iri = 'http://www.emmc.info/emmc-csa/demo#'


# Generate metadata and store it in a JSON file
#
# This does not include all of EMMO, but only the new classes,
# bonded_atom and all classes that these relates to.
classes = list(onto.classes()) + [
    onto.BondedAtom, onto.Integer, onto.Real, onto.String]
e = EMMO2Meta(ontology=onto, classes=classes, collid='usercase_ontology')
e.save('json', 'usercase_metadata.json', 'mode=w')

print('Generated metadata for the usercase ontology:')
print('  %d instances' % e.coll.count())
print('  %d relations' % len(list(e.coll.relations())))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Plots the user case ontology created with the script `define_ontology.py`.
"""
from emmo import World


# Load usercase ontology from sqlite3 database
world = World(filename='demo.sqlite3')
onto = world.get_ontology('http://www.emmc.info/emmc-csa/demo#')
onto.load()

classes = set(onto.classes())


# Visualise new classes (classes from emmo gets a red border)
graph = onto.get_graph(entities=classes, relations='all',
                       edgelabels=False, addnodes=True, addconstructs=False,
                       excluded_nodes=['SquareLengthDimension'],
                       graph_attr={'rankdir': 'RL'})
graph.add_legend()
graph.save('demo.svg')


# Visualise units and quantities
#entities = [c for c in classes
#            if issubclass(c, (onto.MeasurementUnit, onto.Quantity))]
#graph = onto.get_graph(entities=entities, relations='all',
#                       edgelabels=False, addnodes=True, addconstructs=False,
#                       graph_attr={'rankdir': 'RL'})
#graph.add_legend()
#graph.save('units+properties.svg')


# Visualise Material branch
graph = onto.get_graph(root=onto.Material, relations='all',
                       edgelabels=False, addnodes=False, addconstructs=False)
graph.add_legend()
graph.save('materials.svg')


# Material and properties
entities = [c for c in classes if issubclass(c, onto.Material)]
graph = onto.get_graph(entities=entities, relations='all',
                       edgelabels=False, addnodes=True, addconstructs=False,
                       graph_attr={'rankdir': 'RL'})
graph.add_legend()
graph.save('materials+properties.svg')



#!/usr/bin/env python3
"""Module for representing an EMMO-based ontology, as a collection of
DLite metadata entities.

Entities in the ontology are mapped to DLite as follows:
  - owl class (except EMMO property, see below) -> metadata entities
  - owl `hasProperty` restrictions are interpreted. The "object"
    entity of the relation is added as a SOFT property to the
    "subject" entity.
  - all other owl restriction -> entity + relation(s)
  - owl object property -> relations
  - owl class construct -> entity + relation(s)

TODO:
  - map restriction cardinality to collection diminsions
"""
import sys
import re

import dlite
from dlite import Instance, Dimension, Property
from emmo import get_ontology
from emmo.utils import asstring
import owlready2


# We want an ordered dict.  From Python 3.7 (unofficially 3.6) the
# built-in dict conserves insertion order.
if sys.version_info >= (3, 7):
    odict = dict
else:
    from collections import OrderedDict as odict  # noqa: F401


class EMMO2Meta:
    """A class for representing EMMO or an EMMO-based ontology as a
    collection of metadata entities using DLite.

    Parameters
    ----------
    ontology : string
        URI or path to the ontology to represent.  Defaults to EMMO.
    classes : sequence
        The classes to include.  May be given either as a sequence of
        strings or a sequence of owlready2 classes.  The default is to
        include all of the ontology.
    version : string
        Default version for classes lacking a `version` annotation.
    collid : string
        Set an explicit id to the generated collection.

    Notes
    -----
    The collection UUID is accessable via the `coll.uuid` attribute.
    Use the `collid` argument to provide a human readable id to make
    it easier to later retrieve it from a storage (without having
    to remember its UUID).
    """
    def __init__(self, ontology=None, classes=None, version='0.1',
                 collid=None):
        if ontology is None:
            self.onto = get_ontology()
            self.onto.load()
        elif isinstance(ontology, str):
            self.onto = get_ontology(ontology)
            self.onto.load()
        else:
            self.onto = ontology
        self.version = version
        self.iri = self.onto.base_iri
        self.namespace = self.onto.base_iri.rstrip('#')
        self.coll = dlite.Collection(collid)

        if classes is None:
            classes = self.onto.classes()
        elif isinstance(classes, str):
            classes = [classes]

        for cls in classes:
            self.add_class(cls)

    def get_subclasses(self, cls):
        """Returns a generator yielding all subclasses of owl class `cls`."""
        yield cls
        for subcls in cls.subclasses():
            yield from self.get_subclasses(subcls)

    def get_uri(self, name, version=None):
        """Returns uri (namespace/version/name)."""
        if version is None:
            version = self.version
        return '%s/%s/%s' % (self.namespace, version, name)

    def get_uuid(self, uri=None):
        """Returns a UUID corresponding to `uri`.  If `uri` is None,
        a random UUID is returned."""
        return dlite.get_uuid(uri)

    def get_label(self, entity):
        """Returns a label for entity."""
        if hasattr(entity, 'label'):
            return entity.label.first()
        name = repr(entity)
        label, n = re.subn(r'emmo(-[a-z]+)?\.', '', name)
        return label

    def find_label(self, inst):
        """Returns label for class instance `inst` already added to the
        collection."""
        if hasattr(inst, 'uuid'):
            uuid = inst.uuid
        else:
            uuid = dlite.get_uuid(inst)
        rel = self.coll.find_first(p='_has-uuid', o=uuid)
        if not rel:
            raise ValueError('no class instance with UUID: %s' % uuid)
        return rel.s

    def add(self, entity):
        """Adds owl entity to collection and returns a reference to the
        new metadata."""
        if entity == owlready2.Thing:
            raise ValueError("invalid entity: %s" % entity)
        elif isinstance(entity, owlready2.ThingClass):
            return self.add_class(entity)
        elif isinstance(entity, owlready2.ClassConstruct):
            return self.add_class_construct(entity)
        else:
            raise ValueError("invalid entity: %s" % entity)

    def add_class(self, cls):
        """Adds owl class `cls` to collection and returns a reference to
        the new metadata."""
        if isinstance(cls, str):
            cls = self.onto[cls]
        label = cls.label.first()
        if not self.coll.has(label):
            uri = self.get_uri(label)
            dims, props = self.get_properties(cls)
            e = Instance(uri, dims, props, self.get_description(cls))
            self.coll.add(label, e)
            for r in cls.is_a:
                if r is owlready2.Thing:
                    pass
                elif isinstance(r, owlready2.ThingClass):
                    self.coll.add_relation(label, "is_a", r.label.first())
                    self.add_class(r)
                elif isinstance(r, owlready2.Restriction):
                    # Correct this test if EMMO reintroduce isPropertyOf
                    if (isinstance(r.value, owlready2.ThingClass) and
                            isinstance(r.value, self.onto.Property) and
                            issubclass(r.property, self.onto.hasProperty)):
                        self.add_class(r.value)
                    else:
                        self.add_restriction(r)
                elif isinstance(r, owlready2.ClassConstruct):
                    self.add_class_construct(r)
                else:
                    raise TypeError('Unexpected is_a member: %s' % type(r))
        return self.coll.get(label)

    def get_properties(self, cls):
        """Returns two lists with the dlite dimensions and properties
        correspinding to owl class `cls`."""
        dims = []
        props = []
        dimindices = {}
        propnames = set()
        types = dict(Integer='int', Real='double', String='string')

        def get_dim(r, name, descr=None):
            """Returns dimension index corresponding to dimension name `name`
            for property `r.value`."""
            t = owlready2.class_construct._restriction_type_2_label[r.type]
            if (t in ('some', 'only', 'min') or
                    (t in ('max', 'exactly') and r.cardinality > 1)):
                if name not in dimindices:
                    dimindices[name] = len(dims)
                    dims.append(Dimension(name, descr))
                return [dimindices[name]]
            else:
                return []

        for c in cls.mro():
            if not isinstance(c, owlready2.ThingClass):
                continue
            for r in c.is_a:
                # Note that EMMO currently does not define an inverse for
                # hasProperty.  If we reintroduce that, we should replace
                #
                #     not isinstance(r.property, Inverse) and
                #     issubclass(r.property, self.onto.hasProperty)
                #
                # with
                #
                #     ((isinstance(r.property, Inverse) and
                #       issubclass(Inverse(r.property), onto.isPropertyFor)) or
                #      issubclass(r.property, self.onto.hasProperty))
                #
                if (isinstance(r, owlready2.Restriction) and
                        not isinstance(r.property, owlready2.Inverse) and
                        issubclass(r.property, self.onto.hasProperty) and
                        isinstance(r.value, owlready2.ThingClass) and
                        isinstance(r.value, self.onto.Property)):
                    name = self.get_label(r.value)
                    if name in propnames:
                        continue
                    propnames.add(name)

                    # Default type, ndims and unit
                    if isinstance(r.value, (self.onto.DescriptiveProperty,
                                            self.onto.QualitativeProperty,
                                            self.onto.SubjectiveProperty)):
                        ptype = 'string'
                    else:
                        ptype = 'double'
                    d = []
                    d.extend(get_dim(r, 'n_%ss' % name, 'Number of %s.' %
                                     name))
                    unit = None

                    # Update type, ndims and unit from relations
                    for r2 in [r] + r.value.is_a:
                        if isinstance(r2, owlready2.Restriction):
                            if issubclass(r2.property, self.onto.hasType):
                                typelabel = self.get_label(r2.value)
                                ptype = types[typelabel]
                                d.extend(get_dim(r2, '%s_length' % name,
                                                 'Length of %s' % name))
                            elif issubclass(r2.property, self.onto.hasUnit):
                                unit = self.get_label(r2.value)

                    descr = self.get_description(r.value)
                    props.append(Property(name, type=ptype, dims=d,
                                          unit=unit, description=descr))
        return dims, props

    def add_restriction(self, r):
        """Adds owl restriction `r` to collection and returns a reference
        to it."""
        rtype = owlready2.class_construct._restriction_type_2_label[r.type]
        cardinality = r.cardinality if r.cardinality else 0
        e = self.add_restriction_entity()
        inst = e()
        inst.type = rtype
        inst.cardinality = cardinality
        label = inst.uuid
        vlabel = self.get_label(r.value)
        self.coll.add(label, inst)
        self.coll.add_relation(label, asstring(r.property), vlabel)
        if not self.coll.has(vlabel):
            self.add(r.value)
        return inst

    def add_restriction_entity(self):
        """Adds restriction metadata to collection and returns a reference
        to it."""
        uri = self.get_uri("Restriction")
        if not self.coll.has('Restriction'):
            props = [
                Property('type', type='string', description='Type of '
                         'restriction.  Valid values for `type` are: '
                         '"only", "some", "exact", "min" and "max".'),
                Property('cardinality', type='int', description='The '
                         'cardinality.  Unused for "only" and '
                         '"some" restrictions.'),
            ]
            e = Instance(uri, [], props,
                         "Class restriction.  For each instance of a class "
                         "restriction there should be a relation\n"
                         "\n"
                         "    (r.label, r.property, r.value.label)\n"
                         "\n"
                         "where `r.label` is the label associated with the "
                         "restriction, `r.property` is a relation and "
                         "`r.value.label` is the label of the value of the "
                         "restriction.")
            self.coll.add('Restriction', e)
        return self.coll.get('Restriction')

    def add_class_construct(self, c):
        """Adds owl class construct `c` to collection and returns a reference
        to it."""
        ctype = c.__class__.__name__
        e = self.add_class_construct_entity()
        inst = e()
        label = inst.uuid
        inst.type = ctype
        if isinstance(c, owlready2.LogicalClassConstruct):
            args = c.Classes
        else:
            args = [c.Class]
        for arg in args:
            self.coll.add_relation(label, 'has_argument', self.get_label(arg))
        self.coll.add(label, inst)
        return inst

    def add_class_construct_entity(self):
        """Adds class construct metadata to collection and returns a reference
        to it."""
        uri = self.get_uri("ClassConstruct")
        if not self.coll.has('ClassConstruct'):
            props = [
                Property('type', type='string', description='Type of '
                         'class construct.  Valid values for `type` are: '
                         '"not", "inverse", "and" or "or".'),
            ]
            e = Instance(uri, [], props,
                         "Class construct.  For each instance of a class "
                         "construct there should be one or more relations "
                         "of type\n"
                         "\n"
                         "    (c.label, \"has_argument\", c.value.label)\n"
                         "\n"
                         "where `c.label` is the label associated with the "
                         "class construct, `c.value.label` is the label of "
                         "an argument.")
            self.coll.add('ClassConstruct', e)
        return self.coll.get('ClassConstruct')

    def get_description(self, cls):
        """Returns description for OWL class `cls` by combining its
        annotations."""
        if isinstance(cls, str):
            cls = onto[cls]
        descr = []
        annotations = self.onto.get_annotations(cls)
        if 'definition' in annotations:
            descr.extend(annotations['definition'])
        if 'elucication' in annotations and annotations['elucidation']:
            for e in annotations['elucidation']:
                descr.extend(['', 'ELUCIDATION:', e])
        if 'axiom' in annotations and annotations['axiom']:
            for e in annotations['axiom']:
                descr.extend(['', 'AXIOM:', e])
        if 'comment' in annotations and annotations['comment']:
            for e in annotations['comment']:
                descr.extend(['', 'COMMENT:', e])
        if 'example' in annotations and annotations['example']:
            for e in annotations['example']:
                descr.extend(['', 'EXAMPLE:', e])
        return '\n'.join(descr).strip()

    def save(self, *args, **kw):
        """Saves collection to storage."""
        self.coll.save(*args, **kw)


def main():
    e = EMMO2Meta()
    e.save('json', 'emmo2meta.json', 'mode=w')
    return e


if __name__ == '__main__':
    e = main()
    coll = e.coll
    onto = e.onto



#!/usr/bin/env python3
import sys
import os
import itertools

# Add emmo to sys path
thisdir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(1, os.path.abspath(os.path.join(thisdir, '..', '..')))
from emmo import get_ontology  # noqa: E402, F401

import owlready2  # noqa: E402, F401


emmo = get_ontology()
emmo.load()
# emmo.sync_reasoner()


onto = get_ontology('onto.owl')
onto.imported_ontologies.append(emmo)
onto.base_iri = 'http://emmo.info/examples/test#'

with onto:

    class Hydrogen(emmo.Atom):
        pass

    class Oxygen(emmo.Atom):
        pass

    class H2O(emmo.Molecule):
        """Water molecule."""
        emmo.hasSpatialDirectPart.exactly(2, Hydrogen)
        emmo.hasSpatialDirectPart.exactly(1, Oxygen)

    # Create some
    H1 = Hydrogen()
    H2 = Hydrogen()
    O = Oxygen()  # noqa: E741
    w = H2O()
    w.hasSpatialDirectPart = [H1, H2,  O]


onto.sync_attributes(name_policy='sequential', name_prefix='myonto_')
assert 'myonto_0' in onto
assert 'myonto_6' in onto

onto.sync_attributes(name_policy='uuid', name_prefix='onto_')
assert w.name.startswith('onto_')
assert len(w.name) == 5 + 36


# Remove all traces of onto such that they do not mess up other tests
# when running pytest
for e in itertools.chain(onto.classes(), onto.individuals()):
    owlready2.destroy_entity(e)



# -*- coding: utf-8 -*-
"""
A module adding graphing functionality to emmo.ontology
"""
#
# This module was written before I had a good understanding of DL.
# Should be simplified and improved:
#   - Rewrite OntoGraph to be a standalone class instead of as an mixin
#     for Ontology.
#   - Make it possible to have different styles for different types of
#     relations (by default differentiate between is_a, has_part,
#     has_subdimension, has_sign and has_member).
#   - Factor out methods for finding node trees (may go into Ontology).
#   - Factor out methods to finding relation triplets ``(node1, relation,
#     node2)`` (may go into Ontology).
#   - Consider to switch to graphviz Python package since it seems to have
#     very useful interface to Jupyter Notebook and Qt Console integration,
#     see https://pypi.org/project/graphviz/.
#
import os
import re
import warnings
import tempfile
import xml.etree.ElementTree as ET

import owlready2

from .utils import asstring


def getlabel(e):
    """Returns the label of entity `e`."""
    if hasattr(e, 'prefLabel'):
        return e.prefLabel.first()
    elif hasattr(e, '__name__'):
        return e.__name__
    elif hasattr(e, 'name'):
        return str(e.name)
    else:
        return repr(e)


class OntoGraph:
    """A mixin class used by emmo.ontology.Ontology that adds
    functionality for generating graph representations of the ontology.
    """
    _default_style = {
        'graph': {
            'graph_type': 'digraph', 'rankdir': 'RL', 'fontsize': 8,
            # 'fontname': 'Bitstream Vera Sans', 'splines': 'ortho',
            # 'engine': 'neato',
        },
        'class': {
            'style': 'filled',
            'fillcolor': '#ffffcc',
        },
        'defined_class': {
            'style': 'filled',
            'fillcolor': '#ffc880',
        },
        'individuals': {},
        'is_a': {'arrowhead': 'empty'},
        'equivalent_to': {'color': 'green3', },
        'disjoint_with': {'color': 'red', },
        'inverse_of': {'color': 'orange', },
        # 'other': {'color': 'blue', },
        'relations': {
            'enclosing': {'color': 'red', 'arrowtail': 'diamond',
                          'dir': 'back'},
            'has_subdimension': {'color': 'red', 'arrowtail': 'diamond',
                                 'dir': 'back', 'style': 'dashed'},
            'has_sign': {'color': 'blue', 'style': 'dotted'},
            'has_property': {'color': 'blue'},
            'has_unit': {'color': 'magenta'},
            'has_type': {'color': 'forestgreen'},
        },
        'other': {'color': 'olivedrab'},
        }

    _uml_style = {
        'graph': {
            'graph_type': 'digraph', 'rankdir': 'RL', 'fontsize': 8,
            # 'splines': 'ortho',
        },
        'class': {
            # 'shape': 'record',
            'shape': 'box',
            'fontname': 'Bitstream Vera Sans',
            'style': 'filled',
            'fillcolor': '#ffffe0',
        },
        'defined_class': {
            # 'shape': 'record',
            'shape': 'box',
            'fontname': 'Bitstream Vera Sans',
            'style': 'filled',
            'fillcolor': '#ffc880',
        },
        'individuals': {},
        'is_a': {'arrowhead': 'empty'},
        'equivalent_to': {'color': 'green3'},
        'disjoint_with': {'color': 'red', 'arrowhead': 'none'},
        'inverse_of': {'color': 'orange', 'arrowhead': 'none'},
        # 'other': {'color': 'blue', 'arrowtail': 'diamond', 'dir': 'back'},
        'relations': {
            'enclosing': {'color': 'red', 'arrowtail': 'diamond',
                          'dir': 'back'},
            'has_subdimension': {'color': 'red', 'arrowtail': 'diamond',
                                 'dir': 'back', 'style': 'dashed'},
            'has_sign': {'color': 'blue', 'style': 'dotted'},
            'has_property': {'color': 'blue'},
            'has_unit': {'color': 'magenta'},
            'has_type': {'color': 'forestgreen'},
        },
        'other': {'color': 'blue'},
    }

    def get_dot_graph(self, root=None, graph=None, relations='is_a',
                      leafs=None, parents=False, style=None,
                      edgelabels=True, constraint=False):
        """Returns a pydot graph object for visualising the ontology.

        Parameters
        ----------
        root : None | string | owlready2.ThingClass instance
            Name or owlready2 entity of root node to plot subgraph
            below.  If `root` is None, all classes will be included in the
            subgraph.
        graph : None | pydot.Dot instance
            Pydot graph object to plot into.  If None, a new graph object
            is created using the keyword arguments.
        relations : True | str | sequence
            Sequence of relations to visualise.  If True, all relations are
            included.
        leafs : None | sequence
            A sequence of leaf node names for generating sub-graphs.
        parents : bool | str
            Whether to include parent nodes.  If `parents` is a string,
            only parent nodes down to the given name will included.
        style : None | dict | "uml"
            A dict mapping the name of the different graphical elements
            to dicts of pydot style settings. Supported graphical elements
            include:
              - graph : overall settings pydot graph
              - class : nodes for classes
              - individual : nodes for invididuals
              - is_a : edges for is_a relations
              - equivalent_to : edges for equivalent_to relations
              - disjoint_with : edges for disjoint_with relations
              - inverse_of : edges for inverse_of relations
              - relations : with relation names
                  XXX
              - other : edges for other relations and restrictions
            If style is None, a very simple default style is used.
            Some pre-defined styles can be selected by name (currently
            only "uml").
        edgelabels : bool | dict
            Whether to add labels to the edges of the generated graph.
            It is also possible to provide a dict mapping the
            full labels (with cardinality stripped off for restrictions)
            to some abbriviations.
        constraint : None | bool

        Note: This method requires pydot.
        """

        warnings.warn(
            """The emmo.ontology.get_dot_graph() method is deprecated.
            Use emmo.ontology.get_graph() instead.

            This requires that you install graphviz instead of the old
            pydot package.""", DeprecationWarning)

        # FIXME - double inheritance leads to dublicated nodes. Make sure
        #         to only add a node once!
        import pydot
        from .ontology import NoSuchLabelError
        
        if style is None or style == 'default':
            style = self._default_style
        elif style == 'uml':
            style = self._uml_style
        graph = self._get_dot_graph(root=root, graph=graph,
                                    relations=relations, leafs=leafs,
                                    style=style, edgelabels=edgelabels)
        # Add parents
        # FIXME - facture out into an recursive function to support
        #         multiple inheritance

        if parents and root:
            r = self.get_by_label(root) if isinstance(root, str) else root
            while True:
                parent = r.is_a.first()
                if (parent is None or parent is owlready2.Thing):
                    break
                label = getlabel(parent)
                if self.is_defined(label):
                    node = pydot.Node(label, **style.get('defined_class', {}))
                    # If label contains a hyphen, the node name will
                    # be quoted (bug in pydot?).  To work around, set
                    # the name explicitly...
                    node.set_name(label)
                else:
                    node = pydot.Node(label, **style.get('class', {}))
                    node.set_name(label)
                graph.add_node(node)
                if relations is True or 'is_a' in relations:
                    kw = style.get('is_a', {}).copy()
                    if isinstance(edgelabels, dict):
                        kw['label'] = edgelabels.get('is_a', 'is_a')
                    elif edgelabels:
                        kw['label'] = 'is_a'

                    rootnode = graph.get_node(getlabel(r))[0]
                    edge = pydot.Edge(rootnode, node, **kw)
                    graph.add_edge(edge)
                if (isinstance(parents, str) and label == parents):
                    break
                r = parent
        # Add edges
        for node in graph.get_nodes():
            try:
                entity = self.get_by_label(node.get_name())
            except (KeyError, NoSuchLabelError):
                continue
            # Add is_a edges
            targets = [e for e in entity.is_a if not isinstance(e, (
                owlready2.ThingClass, owlready2.ObjectPropertyClass,
                owlready2.PropertyClass))]

            self._get_dot_add_edges(
                graph, entity, targets, 'relations',
                relations,
                # style=style.get('relations', style.get('other', {})),
                style=style.get('other', {}),
                edgelabels=edgelabels,
                constraint=constraint,
            )

            # Add equivalent_to edges
            if relations is True or 'equivalent_to' in relations:
                self._get_dot_add_edges(
                    graph, entity, entity.equivalent_to, 'equivalent_to',
                    relations, style.get('equivalent_to', {}),
                    edgelabels=edgelabels,
                    constraint=constraint,
                )

            # disjoint_with
            if hasattr(entity, 'disjoints') and (
                    relations is True or 'disjoint_with' in relations):
                self._get_dot_add_edges(
                    graph, entity, entity.disjoints(), 'disjoint_with',
                    relations, style.get('disjoint_with', {}),
                    edgelabels=edgelabels,
                    constraint=constraint,
                )

            # Add inverse_of
            if (hasattr(entity, 'inverse_property') and
                    (relations is True or 'inverse_of' in relations) and
                    entity.inverse_property not in (None, entity)):
                self._get_dot_add_edges(
                    graph, entity, [entity.inverse_property], 'inverse_of',
                    relations, style.get('inverse_of', {}),
                    edgelabels=edgelabels,
                    constraint=constraint,
                )

        return graph

    def _get_dot_add_edges(self, graph, entity, targets, relation,
                           relations, style, edgelabels=True,
                           constraint=None):
        """Adds edges to `graph` for relations between `entity` and all
        members in `targets`.  `style` is a dict with options to pydot.Edge().
        """
        import pydot

        nodes = graph.get_node(getlabel(entity))
        if not nodes:
            return
        node = nodes[0]

        for e in targets:
            s = asstring(e)
            if isinstance(e, owlready2.ThingClass):
                pass
            elif isinstance(e, (owlready2.ObjectPropertyClass,
                                owlready2.PropertyClass)):
                label = getlabel(e)
                nodes = graph.get_node(label)
                if nodes:
                    kw = style.copy()
                    if isinstance(edgelabels, dict):
                        kw['label'] = edgelabels.get(label, label)
                    elif edgelabels:
                        kw['label'] = label
                    edge = pydot.Edge(node, nodes[0], **kw)
                    if constraint is not None:
                        edge.set_constraint(constraint)
                    graph.add_edge(edge)
            elif isinstance(e, owlready2.Restriction):
                rname = getlabel(e.property)
                rtype = owlready2.class_construct._restriction_type_2_label[
                    e.type]

                if relations is True or rname in relations:
                    vname = getlabel(e.value)
                    others = graph.get_node(vname)

                    # Only proceede if there is only one node named `vname`
                    # and an edge to that node does not already exists
                    if (len(others) == 1 and
                            (node.get_name(), vname) not in
                            graph.obj_dict['edges'].keys()):
                        other = others[0]
                    else:
                        continue

                    if rtype in ('min', 'max', 'exactly'):
                        label = '%s %s %d' % (rname, rtype, e.cardinality)
                    else:
                        label = '%s %s' % (rname, rtype)

                    kw = style.copy()
                    if isinstance(edgelabels, dict):
                        slabel = '%s %s' % (rname, rtype)
                        kw['label'] = edgelabels.get(slabel, label) + '   '
                    elif edgelabels:
                        kw['label'] = label + '   '  # Add some extra space

                    edge = pydot.Edge(node, other, **kw)
                    if constraint is not None:
                        edge.set_constraint(constraint)
                    graph.add_edge(edge)

            elif hasattr(self, '_verbose') and self._verbose:
                print('* get_dot_graph() * Ignoring: '
                      '%s %s %s' % (node.get_name(), relation, s))

    def _get_dot_graph(self, root=None, graph=None, relations='is_a',
                       leafs=None, style=None, visited=None,
                       edgelabels=True):
        """Help method. See get_dot_graph(). `visited` is used to filter
        out circular dependencies.
        """
        import pydot

        if graph is None:
            kwargs = style.get('graph', {})
            kwargs.setdefault('newrank', True)
            graph = pydot.Dot(**kwargs)

        if relations is True:
            relations = ['is_a'] + list(self.get_relations())
        elif isinstance(relations, str):
            relations = [relations]
        relations = set(r if isinstance(r, str) else
                        getlabel(r) if len(r.label) == 1 else r.name
                        for r in relations)

        if visited is None:
            visited = set()

        if root is None:
            for root in self.get_root_classes():
                self._get_dot_graph(root=root, graph=graph,
                                    relations=relations, leafs=leafs,
                                    style=style, visited=visited,
                                    edgelabels=edgelabels)
            return graph
        elif isinstance(root, (list, tuple, set)):
            for r in root:
                self._get_dot_graph(root=r, graph=graph,
                                    relations=relations, leafs=leafs,
                                    style=style, visited=visited,
                                    edgelabels=edgelabels)
            return graph
        elif isinstance(root, str):
            root = self.get_by_label(root)

        if root in visited:
            if hasattr(self, '_verbose') and self._verbose:
                warnings.warn('Circular dependency of class %r' %
                              getlabel(root))
            return graph
        visited.add(root)

        label = getlabel(root) #if len(getlabel(root)) == 1 else root.name
        nodes = graph.get_node(label)
        if nodes:
            if len(nodes) > 1:
                warnings.warn(
                    'More than one node corresponding to label: %s' % label)
            node = nodes[0]
        else:
            if self.is_individual(label):
                node = pydot.Node(label, **style.get('individual', {}))
                node.set_name(label)
            elif self.is_defined(label):
                node = pydot.Node(label, **style.get('defined_class', {}))
                node.set_name(label)
            else:
                node = pydot.Node(label, **style.get('class', {}))
                node.set_name(label)
            graph.add_node(node)

        if leafs and label in leafs:
            return graph

        for sc in root.subclasses():
            label = getlabel(sc) 
            if self.is_individual(label):
                subnode = pydot.Node(label, **style.get('individual', {}))
                subnode.set_name(label)
            elif self.is_defined(label):
                subnode = pydot.Node(label, **style.get('defined_class', {}))
                subnode.set_name(label)
            else:
                subnode = pydot.Node(label, **style.get('class', {}))
                subnode.set_name(label)
            graph.add_node(subnode)
            if relations is True or 'is_a' in relations:
                kw = style.get('is_a', {}).copy()
                if isinstance(edgelabels, dict):
                    kw['label'] = edgelabels.get('is_a', 'is_a')
                elif edgelabels:
                    kw['label'] = 'is_a'
                edge = pydot.Edge(subnode, node, **kw)
                graph.add_edge(edge)
            self._get_dot_graph(root=sc, graph=graph,
                                relations=relations, leafs=leafs,
                                style=style, visited=visited,
                                edgelabels=edgelabels)

        return graph

    def get_dot_relations_graph(self, graph=None, relations='is_a',
                                style=None):
        """Returns a disjoined graph of all relations.

        This method simply calls get_dot_graph() with all root relations.
        All arguments are passed on.
        """
        rels = tuple(self.get_relations())
        roots = [relation for relation in rels
                 if not any([r in rels for r in relation.is_a])]
        return self.get_dot_graph(root=roots, graph=graph, relations=relations,
                                  style=style)


def get_figsize(graph):
    """Returns figure size (width, height) in points of figures for the
    current pydot graph object `graph`."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpfile = os.path.join(tmpdir, 'graph.svg')
        graph.write_svg(tmpfile)
        xml = ET.parse(tmpfile)
        svg = xml.getroot()
        width = svg.attrib['width']
        height = svg.attrib['height']
        assert width.endswith('pt')  # ensure that units are in points

    def asfloat(s):
        return float(re.match(r'^[\d.]+', s).group())

    return asfloat(width), asfloat(height)
